import typing

from aerocaps.iges.iges_param import IGESParam


class IGESEntity:

    line_fonts = {
        "no_pattern": 0,
        "solid": 1,
        "dashed": 2,
        "phantom": 3,
        "centerline": 4,
        "dotted": 5,
    }

    color_numbers = {
        "no_color": 0,
        "black": 1,
        "red": 2,
        "green": 3,
        "blue": 4,
        "yellow": 5,
        "magenta": 6,
        "cyan": 7,
        "white": 8,
    }

    def __init__(self, ID: int, parameter_data: typing.List[IGESParam], **kwargs):
        self.entity_ID = IGESParam(ID, "int")
        self.pd_pointer = IGESParam(1, "int")  # First parameter data line for this entity
        self.structure = IGESParam("", "none")
        self.line_font_pattern = IGESParam(self.line_fonts['solid'], "int")
        self.level = IGESParam(1, "int")
        self.view = IGESParam(0, "int")
        self.transformation_matrix = IGESParam(0, "pointer")
        self.label_display_associativity = IGESParam("", "none")
        self.status_number = IGESParam(0, "int")
        self.line_weight_number = IGESParam(105, "int")
        self.color_number = IGESParam(self.color_numbers['cyan'], "int")
        self.parameter_line_count = IGESParam(2, "int")
        self.form_number = IGESParam(0, "int")
        self.reserved = IGESParam("", "none")
        self.entity_label = IGESParam("", "none")
        self.subscript_number = IGESParam(0, "int")
        self.parameter_data = parameter_data
        self.param_delimiter = None
        self.record_delimiter = None
        for k, v in kwargs.items():
            getattr(self, k).value = v

    def write_entity_string(self, entity_starting_line: int, data_starting_line: int, data_string_lines: int):

        entity_string = ""

        def write_line_string(line: typing.List[IGESParam], string_to_write: str):
            for iges_param in line:
                if iges_param.dtype == "none":
                    string_to_write += " " * 8
                elif iges_param.dtype == "int" or iges_param.dtype == "pointer":
                    string_to_write += f"{iges_param.value:8d}"
                else:
                    raise TypeError(f"For an entity, every parameter must have type 'int' or 'pointer' or type 'none'. "
                                    f"Found an "
                                    f"IGESParam with type {type(iges_param)}.")
            return string_to_write

        self.pd_pointer.value = data_starting_line
        self.parameter_line_count.value = data_string_lines

        line1 = [self.entity_ID, self.pd_pointer, self.structure, self.line_font_pattern, self.level, self.view,
                 self.transformation_matrix, self.label_display_associativity, self.status_number]
        line2 = [self.entity_ID, self.line_weight_number, self.color_number, self.parameter_line_count,
                 self.form_number, self.reserved, self.reserved, self.entity_label, self.subscript_number]

        entity_string = write_line_string(line1, entity_string) + f"D{entity_starting_line:7d}\n"

        entity_string = write_line_string(line2, entity_string) + f"D{entity_starting_line + 1:7d}\n"

        return entity_string

    def write_data_string(self, entity_entry_line: int, data_starting_line: int):

        data_string = ""

        current_line = self.entity_ID.write_value_to_python_str() + self.param_delimiter

        for p_idx, p in enumerate(self.parameter_data):
            p_str = p.write_value_to_python_str()
            if len(current_line) + len(p_str) < 64:
                current_line += p_str
                if p_idx < len(self.parameter_data) - 1:
                    current_line += self.param_delimiter
                else:
                    current_line += self.record_delimiter
                    current_line += " " * (64 - len(current_line))
                    current_line += f"{entity_entry_line:8d}P{data_starting_line:7d}\n"
                    data_string += current_line
            else:
                if p_idx < len(self.parameter_data) - 1:
                    current_line += " " * (64 - len(current_line))
                    current_line += f"{entity_entry_line:8d}P{data_starting_line:7d}\n"
                    data_starting_line += 1
                    data_string += current_line
                    current_line = p_str + self.param_delimiter
                else:
                    current_line += " " * (64 - len(current_line))
                    current_line += f"{entity_entry_line:8d}P{data_starting_line:7d}\n"
                    data_starting_line += 1
                    current_line += p_str + self.record_delimiter
                    current_line += " " * (64 - len(p_str + self.record_delimiter))
                    current_line += f"{entity_entry_line:8d}P{data_starting_line:7d}\n"
                    data_starting_line += 1
                    data_string += current_line

        return data_string


class MultiEntityContainer:
    def __init__(self, entities: typing.List[IGESEntity]):
        self.entities = entities

    def write_all_entity_and_data_strings(self):

        full_entity_string = ""
        full_data_string = ""
        data_starting_lines = [1]
        data_string_lengths = []

        # First pass loop to generate the data strings and the data string line numbers
        for entity_idx, entity in enumerate(self.entities):
            entity_entry_line = 1 + 2 * entity_idx
            data_string = entity.write_data_string(entity_entry_line, data_starting_line=data_starting_lines[-1])
            if entity_idx < len(self.entities) - 1:
                data_starting_lines.append(data_starting_lines[-1] + data_string.count('\n'))
            data_string_lengths.append(data_string.count('\n'))

            full_data_string += data_string

        # Second pass loop to use the data string line numbers to generate the entity strings
        for entity_idx, entity in enumerate(self.entities):
            entity_entry_line = 1 + 2 * entity_idx
            full_entity_string += entity.write_entity_string(entity_entry_line, data_starting_lines[entity_idx],
                                                             data_string_lines=data_string_lengths[entity_idx])

        return full_entity_string, full_data_string



from aerocaps.geom.point import Point3D, Origin3D
from aerocaps.units.length import Length


__all__ = [
    "Plane",
    "PlaneX",
    "PlaneY",
    "PlaneZ"
]


class Plane:
    def __init__(self, p0: Point3D, p1: Point3D, p2: Point3D):
        self.p0 = p0
        self.p1 = p1
        self.p2 = p2

    @classmethod
    def plane_parallel_X(cls, distance_from_origin: Length):
        return cls(
            p0=Point3D(x=distance_from_origin, y=Length(m=0.0), z=Length(m=0.0)),
            p1=Point3D(x=distance_from_origin, y=Length(m=1.0), z=Length(m=0.0)),
            p2=Point3D(x=distance_from_origin, y=Length(m=0.0), z=Length(m=1.0))
        )

    @classmethod
    def plane_parallel_Y(cls, distance_from_origin: Length):
        return cls(
            p0=Point3D(x=Length(m=0.0), y=distance_from_origin, z=Length(m=0.0)),
            p1=Point3D(x=Length(m=1.0), y=distance_from_origin, z=Length(m=0.0)),
            p2=Point3D(x=Length(m=0.0), y=distance_from_origin, z=Length(m=1.0))
        )

    @classmethod
    def plane_parallel_Z(cls, distance_from_origin: Length):
        return cls(
            p0=Point3D(x=Length(m=0.0), y=Length(m=0.0), z=distance_from_origin),
            p1=Point3D(x=Length(m=0.0), y=Length(m=1.0), z=distance_from_origin),
            p2=Point3D(x=Length(m=1.0), y=Length(m=0.0), z=distance_from_origin)
        )


class PlaneX(Plane):
    def __init__(self):
        super().__init__(
            p0=Origin3D(),
            p1=Point3D(x=Length(m=0.0), y=Length(m=1.0), z=Length(m=0.0)),
            p2=Point3D(x=Length(m=0.0), y=Length(m=0.0), z=Length(m=1.0))
        )


class PlaneY(Plane):
    def __init__(self):
        super().__init__(
            p0=Origin3D(),
            p1=Point3D(x=Length(m=1.0), y=Length(m=0.0), z=Length(m=0.0)),
            p2=Point3D(x=Length(m=0.0), y=Length(m=0.0), z=Length(m=1.0))
        )


class PlaneZ(Plane):
    def __init__(self):
        super().__init__(
            p0=Origin3D(),
            p1=Point3D(x=Length(m=0.0), y=Length(m=1.0), z=Length(m=0.0)),
            p2=Point3D(x=Length(m=1.0), y=Length(m=0.0), z=Length(m=0.0))
        )



import typing

import numpy as np
import numpy.linalg as la

from aerocaps.geom.vector import Vector3D

__all__ = [
    "Transformation2D",
    "Transformation3D",
    "transform_points_into_coordinate_system",
]


class Transformation2D:
    def __init__(self, tx: list = 0.0, ty: list = 0.0, r: list = 0.0, sx: list = 1.0,
                 sy: list = 1.0, rotation_units: str = 'rad', order='r,s,t'):
        """
        Allows for arbitrary 2D transformations on a set of coordinates of size 2 x N"""

        self.tx = tx
        self.ty = ty
        self.r = r
        self.sx = sx
        self.sy = sy

        if rotation_units == 'rad':
            self.r_rad = self.r
            self.r_deg = np.rad2deg(self.r)
        elif rotation_units == 'deg':
            self.r_deg = self.r
            self.r_rad = np.deg2rad(self.r)

        self.order = order

        self.r_mat = None
        self.s_mat = None
        self.t_mat = None

        self.M = np.eye(3)  # 3 x 3 identity matrix

        self.r_mat = self.generate_rotation_matrix()
        self.s_mat = self.generate_scale_matrix()
        self.t_mat = self.generate_translation_matrix()
        self.generate_transformation_matrix()

    def generate_rotation_matrix(self) -> np.ndarray:
        return np.array([[np.cos(self.r), -np.sin(self.r), 0],
                        [np.sin(self.r), np.cos(self.r), 0],
                        [0, 0, 1]])

    def generate_scale_matrix(self) -> np.ndarray:
        return np.array([[self.sx, 0, 0],
                         [0, self.sy, 0],
                         [0, 0, 1]])

    def generate_translation_matrix(self) -> np.ndarray:
        return np.array([[1, 0, self.tx],
                         [0, 1, self.ty],
                         [0, 0, 1]])

    def generate_transformation_matrix(self):
        for idx, operation in enumerate(self.order.split(',')):
            if operation == 'r':
                self.M = self.M @ self.r_mat.T
            elif operation == 's':
                self.M = self.M @ self.s_mat.T
            elif operation == 't':
                self.M = self.M @ self.t_mat.T
            else:
                raise TransformationError(f'Invalid value \'{operation}\' found in 2-D transformation '
                                          f'(must be \'r\', \'s\', or \'t\'')

    def transform(self, coordinates: np.ndarray):
        r"""
        Computes the transformation of the coordinates.

        Parameters
        ----------
        coordinates: np.ndarray
          Size :math:`N \times 2`, where N is the number of coordinates. The columns represent :math:`x` and :math:`y`.
        """
        return (np.column_stack((coordinates, np.ones(len(coordinates)))) @ self.M)[:, :2]  # x' = x * M


class Transformation3D:
    def __init__(self, tx: float = 0.0, ty: float = 0.0, tz: float = 0.0, rx: float = 0.0, ry: float = 0.0,
                 rz: float = 0.0, sx: float = 1.0, sy: float = 1.0, sz: float = 1.0,
                 rotation_units: str = 'rad', order='rx,ry,rz,s,t'):
        r"""
        Allows for arbitrary 3D transformations on a set of coordinates of size :math:`3 \times N`
        """

        self.tx = tx
        self.ty = ty
        self.tz = tz
        self.rx = rx
        self.ry = ry
        self.rz = rz
        self.sx = sx
        self.sy = sy
        self.sz = sz

        if rotation_units == 'rad':
            self.rx_rad = self.rx
            self.rx_deg = np.rad2deg(self.rx)
            self.ry_rad = self.ry
            self.ry_deg = np.rad2deg(self.ry)
            self.rz_rad = self.rz
            self.rz_deg = np.rad2deg(self.rz)
        elif rotation_units == 'deg':
            self.rx_deg = self.rx
            self.rx_rad = np.deg2rad(self.rx)
            self.ry_deg = self.ry
            self.ry_rad = np.deg2rad(self.ry)
            self.rz_deg = self.rz
            self.rz_rad = np.deg2rad(self.rz)

        self.order = order.split(",")

        self.rx_mat = None
        self.ry_mat = None
        self.rz_mat = None
        self.s_mat = None
        self.t_mat = None

        self.M = np.eye(4)  # 4 x 4 identity matrix

        self.rx_mat = self.generate_rotation_matrix_x()
        self.ry_mat = self.generate_rotation_matrix_y()
        self.rz_mat = self.generate_rotation_matrix_z()
        self.s_mat = self.generate_scale_matrix()
        self.t_mat = self.generate_translation_matrix()
        self.generate_transformation_matrix()

    def generate_rotation_matrix_x(self) -> np.ndarray:
        return np.array([[1, 0, 0, 0],
                         [0, np.cos(self.rx_rad), np.sin(self.rx_rad), 0],
                         [0, -np.sin(self.rx_rad), np.cos(self.rx_rad), 0],
                         [0, 0, 0, 1]])

    def generate_rotation_matrix_y(self) -> np.ndarray:
        return np.array([[np.cos(self.ry_rad), 0, -np.sin(self.ry_rad), 0],
                         [0, 1, 0, 0],
                         [np.sin(self.ry_rad), 0, np.cos(self.ry_rad), 0],
                         [0, 0, 0, 1]])

    def generate_rotation_matrix_z(self) -> np.ndarray:
        return np.array([[np.cos(self.rz_rad), -np.sin(self.rz_rad), 0, 0],
                         [np.sin(self.rz_rad), np.cos(self.rz_rad), 0, 0],
                         [0, 0, 1, 0],
                         [0, 0, 0, 1]])

    def generate_scale_matrix(self) -> np.ndarray:
        return np.array([[self.sx, 0, 0, 0],
                         [0, self.sy, 0, 0],
                         [0, 0, self.sz, 0],
                         [0, 0, 0, 1]])

    def generate_translation_matrix(self) -> np.ndarray:
        return np.array([[1, 0, 0, self.tx],
                         [0, 1, 0, self.ty],
                         [0, 0, 1, self.tz],
                         [0, 0, 0, 1]])

    def generate_transformation_matrix(self):
        for idx, operation in enumerate(self.order):
            if operation == 'rx':
                self.M = self.M @ self.rx_mat.T
            elif operation == 'ry':
                self.M = self.M @ self.ry_mat.T
            elif operation == 'rz':
                self.M = self.M @ self.rz_mat.T
            elif operation == 's':
                self.M = self.M @ self.s_mat.T
            elif operation == 't':
                self.M = self.M @ self.t_mat.T
            else:
                raise TransformationError(f'Invalid value \'{operation}\' found in 3-D transformation order '
                                          f'(must be \'rx\', \'ry\', \'rz\', \'s\', or \'t\'')

    def transform(self, coordinates: np.ndarray):
        """
        Computes the transformation of the coordinates.

        Parameters
        ----------
        coordinates: np.ndarray
          Size N x 3, where N is the number of coordinates. The columns represent :math:`x` and :math:`y`.
        """
        return (np.column_stack((coordinates, np.ones(len(coordinates)))) @ self.M)[:, :3]  # x' = x * M


def _convert_list_of_csys_vectors_to_homogeneous_matrix(csys_vectors: typing.List[Vector3D]) -> np.ndarray:
    r"""
    Converts a list of unit vectors describing a coordinate system to a matrix using homogeneous coordinates.

    Parameters
    ----------
    csys_vectors: typing.List[Vector3D]
        Vectors describing the coordinate system. Should be in the order :math:`(i,j,k)`. The vectors are normalized
        automatically.

    Returns
    -------
    np.ndarray
        Homogeneous matrix of size :math:`4 \times 4`
    """
    mat = np.zeros((4, 4))
    mat[:3, 0] = csys_vectors[0].get_normalized_vector().as_array()
    mat[:3, 1] = csys_vectors[1].get_normalized_vector().as_array()
    mat[:3, 2] = csys_vectors[2].get_normalized_vector().as_array()
    mat[3, 3] = 1.0
    return mat


def transform_points_into_coordinate_system(points_in_current_csys: np.ndarray,
                                            current_csys: typing.List[Vector3D],
                                            primed_csys: typing.List[Vector3D]) -> np.ndarray:
    r"""
    Transforms a set of points from one coordinate system into another.

    Parameters
    ----------
    points_in_current_csys: np.ndarray
        Array of points, :math:`N \times 3`
    current_csys: typing.List[Vector3D]
        List of vector objects describing the coordinate system the points currently reside in. Should be in
        the order :math:`(i,j,k)`. The vectors are normalized automatically by the code.
    primed_csys: typing.List[Vector3D]
        List of vector objects describing the coordinate system into which the points should be transformed.
        Should be in the order :math:`(i,j,k)`. The vectors are normalized automatically by the code.

    Returns
    -------
    np.ndarray
        Transformed points
    """
    mat_current = _convert_list_of_csys_vectors_to_homogeneous_matrix(current_csys)
    mat_prime = _convert_list_of_csys_vectors_to_homogeneous_matrix(primed_csys)
    mat_current_inverse = la.inv(mat_current)
    # homogeneous_transposed_points = np.column_stack((
    #     points_in_current_csys, np.ones(points_in_current_csys.shape[0])
    # )).T
    homogeneous_transposed_points = np.insert(
        points_in_current_csys, 3, 1.0, axis=1
    ).T
    homogeneous_transformed_points = mat_prime @ (mat_current_inverse @ homogeneous_transposed_points)
    return homogeneous_transformed_points.T[:, :3]


class TransformationError(Exception):
    pass



from datetime import datetime

from aerocaps.iges.iges_param import IGESParam
from aerocaps.iges import global_section_col_width


class GlobalParams:
    """Global parameter section setup for the IGES file format containing defaults for each value"""

    units_indicators = {
        "inches": (1, "INCH"),
        "millimeters": (2, "MM"),
        "feet": (4, "FT"),
        "miles": (5, "MI"),
        "meters": (6, "M"),
        "kilometers": (7, "KM"),
        "mils": (8, "MIL"),
        "microns": (9, "UM"),
        "centimeters": (10, "CM"),
        "microinches": (11, "UIN"),
    }

    def __init__(self, units: str):
        if units not in self.units_indicators.keys():
            raise ValueError(f"Unit type must be one of {[k for k in self.units_indicators.keys()]}. Unit type"
                             f"input was {units}")
        self.parameter_delimiter_char = IGESParam(",", "string")
        self.record_delimiter_char = IGESParam(";", "string")
        self.product_id_sender = IGESParam("aerocaps_geometry", "string")
        self.file_name = IGESParam("", "none")
        self._native_system_id = IGESParam(f"aerocaps", "string")
        self._preprocessor_version = IGESParam(f"aerocaps", "string")
        self.binary_bits_int = IGESParam(32, "int")
        self.max_power_sp = IGESParam(38, "int")  # single-precision
        self.sig_digits_sp = IGESParam(16, "int")  # single-precision
        self.max_power_dp = IGESParam(38, "int")  # double-precision
        self.sig_digits_dp = IGESParam(16, "int")  # double-precision
        self.product_id_receiver = IGESParam("aerocaps_geometry", "string")
        self.model_space_scale = IGESParam(1.0, "real")
        self.units_flag = IGESParam(self.units_indicators[units][0], "int")
        self.units_name = IGESParam(self.units_indicators[units][1], "string")
        self.max_line_weight_gradations = IGESParam(300, "int")
        self.width_max_line_weight = IGESParam(2.0, "real")
        self.date_and_time_file_generation = IGESParam(datetime.now(), "datetime")
        self.min_model_resolution = IGESParam(1.0e-5, "real")
        self.approx_max_coord_value = IGESParam(100000.0, "real")
        self.author_name = IGESParam("", "none")  # dtype can also be set to "string"
        self.author_org = IGESParam("", "none")  # dtype can also be set to "string"
        self.spec_compliance_flag = IGESParam(11, "int")
        self.spec_drafting_flag = IGESParam(0, "int")
        self.date_time_last_modification = IGESParam(datetime.now(), "datetime")

    def write_globals_string(self):
        gp_list = [v.write_value_to_python_str() for v in vars(self).values()]
        gp_string = self.parameter_delimiter_char.value.join(gp_list)
        gp_string += self.record_delimiter_char.value
        return self.word_wrap_globals(gp_string)

    @staticmethod
    def word_wrap_globals(old_gp_string: str):
        new_gp_string = ""
        start_idx, end_idx = 0, min(global_section_col_width, len(old_gp_string))
        gp_line_counter = 0
        for wrap_idx in range(int(len(old_gp_string) / global_section_col_width) + 1):
            gp_line_counter += 1
            new_gp_string += old_gp_string[start_idx:end_idx]
            new_gp_string += " " * (global_section_col_width - (end_idx - start_idx))
            new_gp_string += f"G{gp_line_counter:7d}\n"
            start_idx += global_section_col_width
            end_idx += min(global_section_col_width,
                           len(old_gp_string) - (wrap_idx + 1) * global_section_col_width)
        return new_gp_string



import numpy as np

from aerocaps.units.unit import Unit


__all__ = [
    "Angle"
]


angle_conversions_from_rad = {
    'deg': lambda r: np.rad2deg(r),
}

angle_conversions_to_rad = {
    'deg': lambda d: np.deg2rad(d),
}


class Angle(Unit):
    def __init__(self, rad=None, deg=None):
        self._rad, self._deg = None, None
        if rad is not None:
            self.rad = rad
        elif deg is not None:
            self.deg = deg
        super().__init__(primary_unit="rad")

    def set_all(self):
        """After setting the temperature in rad, set the temperature for all other units"""
        for k, v in angle_conversions_from_rad.items():
            setattr(self, f'_{k}', v(self._rad))

    @property
    def rad(self):
        return self._rad

    @rad.setter
    def rad(self, value):
        self._rad = value
        self.set_all()

    @property
    def deg(self):
        return self._deg

    @deg.setter
    def deg(self, value):
        self._rad = angle_conversions_to_rad['deg'](value)
        self.set_all()



import os

import numpy as np
import pyvista as pv

print(os.getcwd())


#from aerocaps import DATA_DIR
from aerocaps.geom.point import Point3D
from aerocaps.geom.surfaces import NURBSSurface, BezierSurface, RationalBezierSurface, SurfaceEdge
from aerocaps.geom.curves import Bezier3D,Line3D
from aerocaps.geom import NegativeWeightError
from aerocaps.units.angle import Angle
from aerocaps.iges.iges_generator import IGESGenerator
from aerocaps import TEST_DIR



def test_nurbs_revolve():
    axis = Line3D(p0=Point3D.from_array(np.array([0.0, 0.0, 0.0])),
                  p1=Point3D.from_array(np.array([0.0, 0.0, 1.0])))
    cubic_bezier_cps = np.array([
        [0.0, -1.0, 0.0],
        [0.0, -1.2, 0.5],
        [0.0, -1.3, 1.0],
        [0.0, -0.8, 1.5]
    ])
    bezier = Bezier3D([Point3D.from_array(p) for p in cubic_bezier_cps])
    nurbs_surface = NURBSSurface.from_bezier_revolve(bezier, axis, Angle(deg=15.0), Angle(deg=130.0))

    iges_entities = [nurbs_surface.to_iges()]
    cp_net_points, cp_net_lines = nurbs_surface.generate_control_point_net()
    iges_entities.extend([cp_net_point.to_iges() for cp_net_point in cp_net_points])
    iges_entities.extend([cp_net_line.to_iges() for cp_net_line in cp_net_lines])

    iges_file = os.path.join(TEST_DIR, "nurbs_test.igs")
    iges_generator = IGESGenerator(iges_entities, "meters")
    iges_generator.generate(iges_file)

    point_array = nurbs_surface.evaluate(30, 30)
    for point in point_array[:, 0, :]:
        radius = np.sqrt(point[0] ** 2 + point[1] ** 2)
        assert np.isclose(radius, 1.0, 1e-10)
    for point in point_array[:, -1, :]:
        radius = np.sqrt(point[0] ** 2 + point[1] ** 2)
        assert np.isclose(radius, 0.8, 1e-10)


def test_bezier_surface_1():
    """
    Tests the continuity enforcement method across many random pairs of 4x4 ``BezierSurface``s.
    """
    # FOR TESTING 4x4 and 4x4 first
    n = 4
    m = 4
    num_samples = 50
    rng = np.random.default_rng(seed=42)

    cp_sets_1 = rng.random((num_samples, n+1, m+1, 3))
    cp_sets_2 = rng.random((num_samples, n+1, m+1, 3))

    #Loop through different sides of the 4x4
    
    for i in range(4):
        for j in range(4):
            side_self=SurfaceEdge(i)
            side_other=SurfaceEdge(j)



            # Loop through each pair of control point meshes
            for cp_set1, cp_set2 in zip(cp_sets_1, cp_sets_2):
                bez_surf_1 = BezierSurface(cp_set1)
                bez_surf_2 = BezierSurface(cp_set2)

                # Enforce G0, G1, and G2 continuity
                bez_surf_1.enforce_g0g1g2(bez_surf_2, 1.0, side_self, side_other)

                # Verify G0, G1, and G2 continuity
                bez_surf_1.verify_g0(bez_surf_2, side_self, side_other)
                bez_surf_1.verify_g1(bez_surf_2, side_self, side_other)
                bez_surf_1.verify_g2(bez_surf_2, side_self, side_other)



def test_bezier_surface_2():
    """
    Tests the continuity enforcement method across many random pairs of randomly sized Bezier Surfaces for the parallel degree verification
    """
    for n in range(50):
        # GENERATE THE control point arrays by randomly making a 3 element array
        random_array = np.random.randint(low=4, high=15, size=3)

        #Pick the control points randomly from the 3 element array. 
        n1 = random_array[np.random.randint(0, len(random_array) )]
        m1 = random_array[np.random.randint(0, len(random_array) )]
        
        n1m1_array=np.array([n1,m1])
        random_value=np.random.randint(0,2)
        if random_value==0:
            n2= n1m1_array[np.random.randint(0,2)]
            m2= random_array[np.random.randint(0, len(random_array) )]
        else:
            m2= n1m1_array[np.random.randint(0,2)]
            n2= random_array[np.random.randint(0, len(random_array) )]



        
        rng = np.random.default_rng(seed=42)

        cp_1 = rng.random(( n1+1, m1+1, 3))
        cp_2 = rng.random(( n2+1, m2+1, 3))

        #Loop through different compatible sides

        if (np.shape(cp_1)[0]==np.shape(cp_2)[0]):
            i_vals=np.array([0,1])
            j_vals=np.array([0,1])

        elif (np.shape(cp_1)[0]==np.shape(cp_2)[1]):
            i_vals=np.array([0,1])
            j_vals=np.array([2,3])

        elif (np.shape(cp_1)[1]==np.shape(cp_2)[0]):
            i_vals=np.array([2,3])
            j_vals=np.array([0,1])
        
        elif (np.shape(cp_1)[1]==np.shape(cp_2)[1]):
            i_vals=np.array([2,3])
            j_vals=np.array([2,3])
        
        else:
            raise ValueError("Could not find matching degrees between the surfaces")
        
        for i in i_vals:
            for j in j_vals:
                side_self=SurfaceEdge(i)
                side_other=SurfaceEdge(j)

                # Loop through each pair of control point meshes
                
                bez_surf_1 = BezierSurface(cp_1)
                bez_surf_2 = BezierSurface(cp_2)

                # Enforce G0, G1, and G2 continuity
                bez_surf_1.enforce_g0g1g2(bez_surf_2, 1.0, side_self, side_other)

                # Verify G0, G1, and G2 continuity
                bez_surf_1.verify_g0(bez_surf_2, side_self, side_other)
                bez_surf_1.verify_g1(bez_surf_2, side_self, side_other)
                bez_surf_1.verify_g2(bez_surf_2, side_self, side_other)



def test_bezier_surface_3():
    """
    Tests the continuity enforcement method across many random pairs of randomly sized Bezier Surfaces for verifying whether the tests raise assertion errors when surfaces are incompatible.
    """
    for n in range(50):
        n1 = np.random.randint(low=4, high=10)
        m1 = np.random.randint(low=4, high=10)
        n2 = np.random.randint(low=4, high=10)
        m2 = np.random.randint(low=4, high=10)
        
        rng = np.random.default_rng(seed=42)

        cp_1 = rng.random(( n1+1, m1+1, 3))
        cp_2 = rng.random(( n2+1, m2+1, 3))

        
        
        for i in range(4):
            for j in range(4):
                side_self=SurfaceEdge(i)
                side_other=SurfaceEdge(j)

                # Loop through each pair of control point meshes
                
                bez_surf_1 = BezierSurface(cp_1)
                bez_surf_2 = BezierSurface(cp_2)
                
                try:
                    # Enforce G0, G1, and G2 continuity
                    bez_surf_1.enforce_g0g1g2(bez_surf_2, 1.0, side_self, side_other)

                    # Verify G0, G1, and G2 continuity
                    bez_surf_1.verify_g0(bez_surf_2, side_self, side_other)
                    bez_surf_1.verify_g1(bez_surf_2, side_self, side_other)
                    bez_surf_1.verify_g2(bez_surf_2, side_self, side_other)
                except AssertionError:
                    continue

def test_Rational_Bezier_Surface_1():
    """
    Tests the continuity enforcement method across many random pairs of 4x4 ``RationalBezierSurface``s.
    """
    rng = np.random.default_rng(seed=42)
    negative_counter=0
    for it in range(50):
        n=rng.integers(4, 9)
        m=n
        #rng = np.random.default_rng(seed=42)

        # cp_1 = np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1]],
        #                  [[0,1,1],[1,1,0],[2,1,1],[3,1,1]],
        #                  [[0,2,0],[1,2,1],[2,2,0],[3,2,1]],
        #                  [[0,3,0],[1,3,1],[2,3,1],[3,3,1]]],dtype=np.float64)  

        cp_1 =rng.random(( n+1, m+1, 3))
                 
        # cp_2 =  np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1]],
        #                  [[0,1,2],[1,1,1],[2,1,1],[3,1,1]],
        #                  [[0,2,0],[1,2,0],[2,2,1],[3,2,1]],
        #                  [[0,3,0],[1,3,1],[2,3,1],[3,3,1]]],dtype=np.float64)            
        cp_2 =rng.random(( n+1, m+1, 3))
        w_1 = rng.uniform(0.4, 0.5, (n+1, m+1))
        w_2 = rng.uniform(0.9, 1.2, (n+1, m+1))

        for i in range(1):
            for j in range(1):
                side_self=SurfaceEdge(i)
                side_other=SurfaceEdge(j)

                # Loop through each pair of control point meshes
                
                Rat_bez_surf_1 = RationalBezierSurface(cp_1,w_1)
                Rat_bez_surf_2 = RationalBezierSurface(cp_2,w_2)
                
                try:
                    Rat_bez_surf_1.enforce_g0g1g2(Rat_bez_surf_2, 1.0, side_self, side_other)
                    Rat_bez_surf_1.verify_g0(Rat_bez_surf_2, side_self, side_other)

                    # g1_self=Rat_bez_surf_1.get_first_derivs_along_edge(side_self)
                    # g2_self=Rat_bez_surf_2.get_first_derivs_along_edge(side_other)

                    # g1_self_v2=Rat_bez_surf_1.get_first_derivs_along_edge_v2(side_self)
                    # g2_self_v2=Rat_bez_surf_2.get_first_derivs_along_edge_v2(side_other)
                    
                    #print(f"{g1_self=},{g2_self=}")
                    #print(f"{g1_self_v2=},{g2_self_v2=}")
                    Rat_bez_surf_1.verify_g1(Rat_bez_surf_2, side_self, side_other)
                    Rat_bez_surf_1.verify_g2(Rat_bez_surf_2, side_self, side_other)
                except NegativeWeightError:
                    negative_counter+=1
                    #print(f'{it=},{negative_counter=}')

                    

                #except NegativeWeightError:
                #print(f"{negative_counter=}")
                #negative_counter+=1
                #continue
                # Enforce G0, G1, and G2 continuity

                # Verify G0, G1, and G2 continuity
                
                
                
    #print(f"{negative_counter=}")


def test_Rational_Bezier_Surface_2():
    """
    Tests the continuity enforcement method across many random pairs of 4x4 ``RationalBezierSurface``s.
    """
    rng = np.random.default_rng(seed=42)
    Assertion_error_counter=0
    Negative_error_counter=0
    for n in range(20):

        random_array = rng.integers(low=3, high=5, size=3)

        #Pick the control points randomly from the 3 element array. 
        n1 = random_array[rng.integers(0, len(random_array) )]
        m1 = random_array[rng.integers(0, len(random_array) )]
        
        n1m1_array=np.array([n1,m1])
        random_value=rng.integers(0,2)
        if random_value==0:
            n2= n1m1_array[rng.integers(0,2)]
            m2= random_array[rng.integers(0, len(random_array) )]
        else:
            m2= n1m1_array[rng.integers(0,2)]
            n2= random_array[rng.integers(0, len(random_array) )]



        
        

        cp_1 = rng.random(( n1+1, m1+1, 3))

        # cp_1 = np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1]],
        #                  [[0,1,1],[1,1,0],[2,1,1],[3,1,1]],
        #                  [[0,2,0],[1,2,1],[2,2,0],[3,2,1]],
        #                  [[0,3,0],[1,3,1],[2,3,1],[3,3,1]]],dtype=np.float64)  
        cp_2 = rng.random(( n2+1, m2+1, 3))

        # cp_2 =  np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1]],
        #                  [[0,1,2],[1,1,1],[2,1,1],[3,1,1]],
        #                  [[0,2,0],[1,2,0],[2,2,1],[3,2,1]],
        #                  [[0,3,0],[1,3,1],[2,3,1],[3,3,1]]],dtype=np.float64)      

        w_1 = rng.uniform(0.4, 0.5, (n1+1, m1+1))
        w_2 = rng.uniform(0.9, 1.2, (n2+1, m2+1))

        #Loop through different compatible sides

        if (np.shape(cp_1)[0]==np.shape(cp_2)[0]):
            i_vals=np.array([0,1])
            j_vals=np.array([0,1])

        elif (np.shape(cp_1)[0]==np.shape(cp_2)[1]):
            i_vals=np.array([0,1])
            j_vals=np.array([2,3])

        elif (np.shape(cp_1)[1]==np.shape(cp_2)[0]):
            i_vals=np.array([2,3])
            j_vals=np.array([0,1])
        
        elif (np.shape(cp_1)[1]==np.shape(cp_2)[1]):
            i_vals=np.array([2,3])
            j_vals=np.array([2,3])
        
        else:
            raise ValueError("Could not find matching degrees between the surfaces")
        
        for i in i_vals:
            for j in j_vals:
                side_self=SurfaceEdge(i)
                side_other=SurfaceEdge(j)

                # Loop through each pair of control point meshes
                
                Rat_bez_surf_1 = RationalBezierSurface(cp_1,w_1)
                Rat_bez_surf_2 = RationalBezierSurface(cp_2,w_2)

                # Enforce G0, G1, and G2 continuity
                try:
                    Rat_bez_surf_1.enforce_g0g1g2(Rat_bez_surf_2, 1.0, side_self, side_other)
                    
                    # Verify G0, G1, and G2 continuity
                    Rat_bez_surf_1.verify_g0(Rat_bez_surf_2, side_self, side_other)
                    Rat_bez_surf_1.verify_g1(Rat_bez_surf_2, side_self, side_other)
                    Rat_bez_surf_1.verify_g2(Rat_bez_surf_2, side_self, side_other)
                except AssertionError:
                    Assertion_error_counter+=1
                    
                except NegativeWeightError:
                    Negative_error_counter+=1
    print(f'{n=},{Assertion_error_counter=}')
    print(f'{n=},{Negative_error_counter=}')


test_Rational_Bezier_Surface_2()

def test_Rational_Bezier_Surface_3():
    """
    Tests the continuity enforcement method across many random pairs of 4x4 ``RationalBezierSurface``s.
    """
    rng = np.random.default_rng(seed=42)
    for n in range(1):

        random_array = rng.integers(low=4, high=15, size=3)

        #Pick the control points randomly from the 3 element array. 
        n1 = random_array[rng.integers(0, len(random_array) )]
        m1 = random_array[rng.integers(0, len(random_array) )]
        
        n1m1_array=np.array([n1,m1])
        random_value=rng.integers(0,2)
        if random_value==0:
            n2= n1m1_array[rng.integers(0,2)]
            m2= random_array[rng.integers(0, len(random_array) )]
        else:
            m2= n1m1_array[rng.integers(0,2)]
            n2= random_array[rng.integers(0, len(random_array) )]



        
        

        #cp_1 = rng.random(( n1+1, m1+1, 3))

        cp_1 = np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1]],
                         [[0,1,1],[1,1,0],[2,1,1],[3,1,1]],
                         [[0,2,0],[1,2,1],[2,2,0],[3,2,1]],
                         [[0,3,0],[1,3,1],[2,3,1],[3,3,1]]],dtype=np.float64)  
        
        #cp_2 = rng.random(( n2+1, m2+1, 3))

        cp_2 =  np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1]],
                         [[0,1,2],[1,1,1],[2,1,1],[3,1,1]],
                         [[0,2,0],[1,2,0],[2,2,1],[3,2,1]],
                         [[0,3,0],[1,3,1],[2,3,1],[3,3,1]]],dtype=np.float64)            

        w_1 = rng.uniform(0.9, 1.1, (n1+1, m1+1))
        w_2 = rng.uniform(0.9, 1.1, (n2+1, m2+1))

        #Loop through different compatible sides

        if (np.shape(cp_1)[0]==np.shape(cp_2)[0]):
            i_vals=np.array([0,1])
            j_vals=np.array([0,1])

        elif (np.shape(cp_1)[0]==np.shape(cp_2)[1]):
            i_vals=np.array([0,1])
            j_vals=np.array([2,3])

        elif (np.shape(cp_1)[1]==np.shape(cp_2)[0]):
            i_vals=np.array([2,3])
            j_vals=np.array([0,1])
        
        elif (np.shape(cp_1)[1]==np.shape(cp_2)[1]):
            i_vals=np.array([2,3])
            j_vals=np.array([2,3])
        
        else:
            raise ValueError("Could not find matching degrees between the surfaces")
        
        for i in i_vals:
            for j in j_vals:
                side_self=SurfaceEdge(i)
                side_other=SurfaceEdge(j)

                # Loop through each pair of control point meshes
                
                Rat_bez_surf_1 = RationalBezierSurface(cp_1,w_1)
                Rat_bez_surf_2 = RationalBezierSurface(cp_2,w_2)

                # Enforce G0, G1, and G2 continuity
                Rat_bez_surf_1.enforce_g0g1g2(Rat_bez_surf_2, 1.0, side_self, side_other)
                
                # Verify G0, G1, and G2 continuity
                Rat_bez_surf_1.verify_g0(Rat_bez_surf_2, side_self, side_other)
                Rat_bez_surf_1.verify_g1(Rat_bez_surf_2, side_self, side_other)
                Rat_bez_surf_1.verify_g2(Rat_bez_surf_2, side_self, side_other)

#test_Rational_Bezier_Surface_3()


####### SCRATCH CODE #######

# iges_entities = [Rat_bez_surf_1.to_iges(),Rat_bez_surf_2.to_iges()]
# cp_net_points, cp_net_lines = Rat_bez_surf_1.generate_control_point_net()
# iges_entities.extend([cp_net_point.to_iges() for cp_net_point in cp_net_points])
# iges_entities.extend([cp_net_line.to_iges() for cp_net_line in cp_net_lines])
# cp_net_points_2, cp_net_lines_2 = Rat_bez_surf_2.generate_control_point_net()
# iges_entities.extend([cp_net_point.to_iges() for cp_net_point in cp_net_points_2])
# iges_entities.extend([cp_net_line.to_iges() for cp_net_line in cp_net_lines_2])

# iges_file = os.path.join(TEST_DIR, "Rat_Bez_test.igs")
# print(f"{iges_file=}")
# iges_generator = IGESGenerator(iges_entities, "meters")
# iges_generator.generate(iges_file)
# print("Generator passed")


#PLOTTER

# plot= pv.plotter()
# Rat_bez_surf_1.plot_surface()
# Rat_bez_surf_1.plot_control_point_mesh_lines(plot)
# Rat_bez_surf_1.plot_control_points(plot)
# Rat_bez_surf_2.plot_surface()
# Rat_bez_surf_2.plot_control_point_mesh_lines(plot)
# Rat_bez_surf_2.plot_control_points(plot)
# plot.show()


__version__ = "0.0.1"


def get_major_project_version() -> str:
    """
    Gets the major project version for AeroCAPS using semantic versioning

    Returns
    -------
    str
        The major project version
    """
    split_version = __version__.split(".")
    return ".".join(split_version[:2])



import numpy as np

import aerocaps
import aerocaps.iges.curves
import aerocaps.iges.surfaces
import aerocaps.iges.iges_generator


def test_fill_surface_xy_plane():
    # Create a triangle in the X-Y plane
    points = np.array([
        [0.0, 0.0, 0.0],
        [1.0, 1.0, 0.0],
        [0.0, 1.0, 0.0]
    ])
    p0 = aerocaps.Point3D.from_array(points[0, :])
    p1 = aerocaps.Point3D.from_array(points[1, :])
    p2 = aerocaps.Point3D.from_array(points[2, :])
    line_1 = aerocaps.Line3D(p0=p0, p1=p1)
    line_2 = aerocaps.Line3D(p0=p1, p1=p2)
    line_3 = aerocaps.Line3D(p0=p2, p1=p0)
    composite = aerocaps.CompositeCurve3D([line_1, line_2, line_3])

    # Create a surface that fully encloses the triangle in the X-Y plane
    corners = np.array([
        [-2.0, -2.0, 0.0],
        [2.0, -2.0, 0.0],
        [2.0, 2.0, 0.0],
        [-2.0, 2.0, 0.0]
    ])
    pa = aerocaps.Point3D.from_array(corners[0, :])
    pb = aerocaps.Point3D.from_array(corners[1, :])
    pc = aerocaps.Point3D.from_array(corners[2, :])
    pd = aerocaps.Point3D.from_array(corners[3, :])
    surf = aerocaps.BezierSurface([[pa, pd], [pb, pc]])

    # Create the parametric space versions of the triangle lines
    parametric_points = np.array([
        [0.5, 0.5, 0.0],
        [0.75, 0.75, 0.0],
        [0.5, 0.75, 0.0]
    ])
    p0_para = aerocaps.Point3D.from_array(parametric_points[0, :])
    p1_para = aerocaps.Point3D.from_array(parametric_points[1, :])
    p2_para = aerocaps.Point3D.from_array(parametric_points[2, :])
    line_1_para = aerocaps.Line3D(p0=p0_para, p1=p1_para)
    line_2_para = aerocaps.Line3D(p0=p1_para, p1=p2_para)
    line_3_para = aerocaps.Line3D(p0=p2_para, p1=p0_para)
    composite_para = aerocaps.CompositeCurve3D([line_1_para, line_2_para, line_3_para])

    # Create the definition for the parametric curve
    curve_on_parametric_surface = aerocaps.CurveOnParametricSurface(
        surf,
        composite_para,
        composite
    )

    # Create the trimmed surface object
    trimmed_surf = aerocaps.TrimmedSurface(surf, curve_on_parametric_surface)

    # Set up the IGES generator and generate the IGES file
    entities = [line.to_iges() for line in [line_1, line_2, line_3]]
    entities.extend([line.to_iges() for line in [line_1_para, line_2_para, line_3_para]])
    entities.append(composite.to_iges(entities[0:3]))
    entities.append(composite_para.to_iges(entities[3:6]))
    entities.append(surf.to_iges())
    entities.append(curve_on_parametric_surface.to_iges(entities[8], entities[7], entities[6]))
    entities.append(trimmed_surf.to_iges(entities[8], entities[9]))
    iges_generator = aerocaps.iges.iges_generator.IGESGenerator(entities, units="meters")
    iges_generator.generate("fill_surface_xy_plane.igs")


def test_fill_surface_xy_plane_new_builtin():
    # Create a triangle in the X-Y plane
    points = np.array([
        [0.0, 0.0, 0.0],
        [1.0, 1.0, 0.0],
        [0.0, 1.0, 0.0]
    ])
    p0 = aerocaps.Point3D.from_array(points[0, :])
    p1 = aerocaps.Point3D.from_array(points[1, :])
    p2 = aerocaps.Point3D.from_array(points[2, :])
    line_1 = aerocaps.Line3D(p0=p0, p1=p1)
    line_2 = aerocaps.Line3D(p0=p1, p1=p2)
    line_3 = aerocaps.Line3D(p0=p2, p1=p0)

    fill = aerocaps.PlanarFillSurfaceCreator([line_1, line_2, line_3])
    iges_generator = aerocaps.iges.iges_generator.IGESGenerator(fill.to_iges(), units="meters")
    iges_generator.generate("fill_surface_xy_plane_new_builtin.igs")



import typing


def flatten_arbitrarily_nested_list_of_lists(xs):
    for x in xs:
        if isinstance(x, typing.Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten_arbitrarily_nested_list_of_lists(x)
        else:
            yield x



import typing

import numpy as np
from numpy.ma.core import inner

from aerocaps.iges.curves import BoundaryCurveIGES, CurveOnParametricSurfaceIGES
from aerocaps.units.angle import Angle
from aerocaps.iges.entity import IGESEntity
from aerocaps.iges.iges_param import IGESParam


class SurfaceOfRevolutionIGES(IGESEntity):
    def __init__(self, axis_of_revolution: IGESEntity, curve: IGESEntity, start_angle: Angle, end_angle: Angle,
                 **entity_kwargs):
        parameter_data = [
            IGESParam(axis_of_revolution, "pointer"),
            IGESParam(curve, "pointer"),
            IGESParam(start_angle.rad, "real"),
            IGESParam(end_angle.rad, "real")
        ]
        super().__init__(120, parameter_data, **entity_kwargs)


class RuledSurfaceIGES(IGESEntity):
    def __init__(self, curve_1: IGESEntity, curve_2: IGESEntity, **entity_kwargs):
        parameter_data = [
            IGESParam(curve_1, "pointer"),
            IGESParam(curve_2, "pointer"),
            IGESParam(0, "int"),
            IGESParam(1, "int")
        ]
        super().__init__(118, parameter_data, **entity_kwargs)


class BoundedSurfaceIGES(IGESEntity):
    def __init__(self,
                 untrimmed_surface: IGESEntity,
                 boundary_curves: typing.List[BoundaryCurveIGES],
                 **entity_kwargs):
        """
        IGES Type 143

        Parameters
        ----------
        untrimmed_surface: IGESEntity
            Surface in which the boundary curves reside
        boundary_curves: typing.List[BoundaryCurveIGES]
            Closed loop of boundary curves that lie on the surface specified by ``untrimmed_surface``
        entity_kwargs
            Additional keyword arguments to pass to the ``IGESEntity`` constructor
        """
        parameter_data = [
            IGESParam(int(any(
                [boundary_curve.parameter_data[0].value == 1 for boundary_curve in boundary_curves])), "int"
            ),
            IGESParam(untrimmed_surface, "pointer"),
            IGESParam(len(boundary_curves), "int"),
            *[IGESParam(boundary_curve, "pointer") for boundary_curve in boundary_curves]
        ]
        super().__init__(143, parameter_data, **entity_kwargs)


class TrimmedSurfaceIGES(IGESEntity):
    def __init__(self,
                 untrimmed_surface: IGESEntity,
                 outer_boundary: CurveOnParametricSurfaceIGES,
                 inner_boundaries: typing.List[CurveOnParametricSurfaceIGES] = None,
                 outer_boundary_is_boundary_of_surface: bool = False,
                 **entity_kwargs
                 ):
        """
        Trimmed Surface, IGES type 144.

        Parameters
        ----------
        untrimmed_surface: IGESEntity
            Surface in which the boundary curves reside
        outer_boundary: CurveOnParametricSurfaceIGES
            Curve on a parametric surface representing the outer boundary for the trimmed surface
        inner_boundaries: typing.List[CurveOnParametricSurfaceIGES]
            List of inner boundary curves
        outer_boundary_is_boundary_of_surface: bool
            Whether the outer boundary represents the actual outer boundary of the untrimmed surface. Default: ``False``
        entity_kwargs
            Additional keyword arguments to pass to the ``IGESEntity`` constructor
        """
        inner_boundaries = [] if inner_boundaries is None else inner_boundaries
        parameter_data = [
            IGESParam(untrimmed_surface, "pointer"),
            IGESParam(int(outer_boundary_is_boundary_of_surface), "int"),
            IGESParam(len(inner_boundaries), "int"),
            IGESParam(outer_boundary, "pointer"),
            *[IGESParam(inner_boundary, "pointer") for inner_boundary in inner_boundaries]
        ]
        super().__init__(144, parameter_data, **entity_kwargs)


class RationalBSplineSurfaceIGES(IGESEntity):
    def __init__(self,
                 control_points: np.ndarray,
                 knots_u: np.ndarray,
                 knots_v: np.ndarray,
                 weights: np.ndarray,
                 degree_u: int, degree_v: int,
                 closed_u: bool = True, closed_v: bool = True,
                 rational: bool = True,
                 periodic_u: bool = False, periodic_v: bool = False,
                 start_u: float = 0.0, end_u: float = 1.0,
                 start_v: float = 0.0, end_v: float = 1.0,
                 **entity_kwargs
                 ):
        assert control_points.ndim == 3
        assert knots_u.ndim == 1
        assert knots_v.ndim == 1
        assert weights.ndim == 2
        self.control_points = control_points
        self.knots_u = knots_u
        self.knots_v = knots_v
        self.weights = weights
        parameter_data = [
            IGESParam(control_points.shape[0] - 1, "int"),
            IGESParam(control_points.shape[1] - 1, "int"),
            IGESParam(degree_u, "int"),
            IGESParam(degree_v, "int"),
            IGESParam(int(closed_u), "int"),
            IGESParam(int(closed_v), "int"),
            IGESParam(int(not rational), "int"),
            IGESParam(int(periodic_u), "int"),
            IGESParam(int(periodic_v), "int"),
            *[IGESParam(k, "real") for k in self.knots_u],
            *[IGESParam(k, "real") for k in self.knots_v],
            *self._flatten_weights(),
            *self._flatten_control_points(),
            IGESParam(start_u, "real"),
            IGESParam(end_u, "real"),
            IGESParam(start_v, "real"),
            IGESParam(end_v, "real")
        ]
        super().__init__(128, parameter_data, **entity_kwargs)

    def _flatten_control_points(self):
        control_point_params_flattened = []
        for j in range(self.control_points.shape[1]):
            for i in range(self.control_points.shape[0]):
                for k in range(self.control_points.shape[2]):
                    control_point_params_flattened.append(IGESParam(self.control_points[i, j, k], "real"))
        return control_point_params_flattened

    def _flatten_weights(self):
        weights_flattened = []
        for j in range(self.weights.shape[1]):
            for i in range(self.weights.shape[0]):
                weights_flattened.append(IGESParam(self.weights[i, j], "real"))
        return weights_flattened


class BezierSurfaceIGES(RationalBSplineSurfaceIGES):
    def __init__(self, control_points: np.ndarray, **entity_kwargs):
        order_u = control_points.shape[0]
        order_v = control_points.shape[1]
        degree_u = order_u - 1
        degree_v = order_v - 1
        knots_u = np.concatenate((np.zeros(order_u), np.ones(order_u)))
        knots_v = np.concatenate((np.zeros(order_v), np.ones(order_v)))
        weights = np.ones(shape=(order_u, order_v))
        super().__init__(control_points=control_points, knots_u=knots_u, knots_v=knots_v, weights=weights,
                         degree_u=degree_u, degree_v=degree_v, **entity_kwargs)



import numpy as np
from scipy.optimize import minimize_scalar

from aerocaps.geom.curves import Bezier3D, PCurve2D, PCurve3D, Line3D
from aerocaps.geom.point import Point3D, Point2D
from aerocaps.geom.transformation import Transformation3D
from aerocaps.geom.vector import Vector3D
from aerocaps.units.angle import Angle


__all__ = [
    "measure_distance_between_points",
    "measure_pitch_angle",
    "measure_distance_point_line",
    "add_vector_to_point",
    "project_point_onto_line",
    "find_t_corresponding_to_minimum_distance_to_point2d",
    "find_t_corresponding_to_minimum_distance_to_point3d",
    "sweep_along_curve",
    "rotate_about_axis",
    "rotate_point_about_axis"
]


def measure_distance_between_points(point_1: Point2D or Point3D or np.ndarray,
                                    point_2: Point2D or Point3D or np.ndarray) -> float:
    if isinstance(point_1, Point2D) or isinstance(point_1, Point3D):
        point_1 = point_1.as_array()
    if isinstance(point_2, Point2D) or isinstance(point_2, Point3D):
        point_2 = point_2.as_array()
    if len(point_1) != len(point_2):
        raise ValueError("Cannot calculate distance between two points of different dimensionality")
    if len(point_1) == 2 and len(point_2) == 2:
        return np.sqrt((point_2[0] - point_1[0]) ** 2 + (point_2[1] - point_1[1]) ** 2)
    elif len(point_1) == 3 and len(point_2) == 3:
        return np.sqrt((point_2[0] - point_1[0]) ** 2 + (point_2[1] - point_1[1]) ** 2 + (point_2[2] - point_1[2]) ** 2)
    else:
        raise ValueError("Points must have dimension 2 or 3")


def measure_pitch_angle(point_1: Point3D or np.ndarray, point_2: Point3D or np.ndarray) -> Angle:
    """
    Translates the two points such that the first point is at the origin and then measures the angle the line
    connecting the two points makes on the X-Z plane
    """
    if isinstance(point_1, Point3D):
        point_1 = point_1.as_array()
    if isinstance(point_2, Point3D):
        point_2 = point_2.as_array()
    point_2_translated = np.array([point_2[0] - point_1[0], point_2[1] - point_1[1], point_2[2] - point_1[2]])
    return Angle(rad=np.arctan2(point_2_translated[2], point_2_translated[0]))


def measure_distance_point_line(point: Point3D, line: Line3D) -> float:
    return measure_distance_between_points(point, project_point_onto_line(point, line))


def add_vector_to_point(vector: Vector3D, point: Point3D) -> Point3D:
    return point + (vector.p1 - vector.p0)


def project_point_onto_line(point: Point3D, line: Line3D) -> Point3D:
    vAB = Vector3D(p0=line.p0, p1=line.p1)
    vAC = Vector3D(p0=line.p0, p1=point)
    vAD_value = vAB.as_array() * (vAB.dot(vAC) / vAB.dot(vAB))
    vAD = Vector3D.from_array(vAD_value)
    return add_vector_to_point(vector=vAD, point=line.p0)


def find_t_corresponding_to_minimum_distance_to_point2d(curve: PCurve2D, point: np.ndarray or Point2D) -> (
        float, float):

    point = point.as_array() if isinstance(point, Point2D) else point

    def minimize_func(t):
        curve_point = curve.evaluate_single_t(t)
        return (curve_point[0] - point[0]) ** 2 + (curve_point[1] - point[1]) ** 2

    res = minimize_scalar(minimize_func, bounds=[0.0, 1.0])
    return res.x, np.sqrt(res.fun)


def find_t_corresponding_to_minimum_distance_to_point3d(curve: PCurve3D, point: np.ndarray or Point3D) -> (
        float, float):

    point = point.as_array() if isinstance(point, Point3D) else point

    def minimize_func(t):
        curve_point = curve.evaluate_single_t(t)
        return (curve_point[0] - point[0]) ** 2 + (curve_point[1] - point[1]) ** 2 + (curve_point[2] - point[2]) ** 2

    res = minimize_scalar(minimize_func, bounds=[0.0, 1.0])
    return res.x, np.sqrt(res.fun)


def sweep_along_curve(primary_curve: Bezier3D, guide_curve: Bezier3D):
    point_list_3d = [primary_curve.control_points]
    for prev_point, current_point in zip(guide_curve.control_points[:-1], guide_curve.control_points[1:]):
        point_list_3d.append([primary_point + current_point - prev_point for primary_point in point_list_3d[-1]])
    return np.array([[[point.x.m, point.y.m, point.z.m] for point in v] for v in point_list_3d])


def rotate_about_axis(points: np.ndarray, axis: Vector3D, angle: Angle) -> np.ndarray:
    ux, uy, uz = axis.normalized_value()
    c = np.cos(angle.rad)
    s = np.sin(angle.rad)
    rotation_matrix = np.array([
        [c + ux**2 * (1 - c), ux * uy * (1 - c) - uz * s, ux * uz * (1 - c) + uy * s],
        [uy * ux * (1 - c) + uz * s, c + uy**2 * (1 - c), uy * uz * (1 - c) - ux * s],
        [uz * ux * (1 - c) - uy * s, uz * uy * (1 - c) + ux * s, c + uz**2 * (1 - c)]
    ])
    return (rotation_matrix @ points.T).T


def rotate_point_about_axis(p: Point3D, ax: Line3D, angle: Angle) -> Point3D:
    reverse_transformation = Transformation3D(tx=-ax.p0.x.m, ty=-ax.p0.y.m, tz=-ax.p0.z.m)
    forward_transformation = Transformation3D(tx=ax.p0.x.m, ty=ax.p0.y.m, tz=ax.p0.z.m)
    p_mat = np.array([p.as_array()])
    p_mat = reverse_transformation.transform(p_mat)
    p_mat = rotate_about_axis(p_mat, ax.get_vector(), angle)
    p_mat = forward_transformation.transform(p_mat)
    return Point3D.from_array(p_mat[0])



import os

import numpy as np
import pyvista as pv

import aerocaps as ac
import aerocaps.examples.bezier_surface


def plot_bezier_surf_2x3(image_dir: str = None):
    r"""
    Plots an :math:`(n \times m)=(2 \times 3)` Bézier surface along with the control point net.

    Parameters
    ----------
    image_dir: str or None
        Where to store the output image. If no location is specified, the image file will not be saved, and the
        plot will be displayed on-screen instead. Default: ``None``
    """
    plot = pv.Plotter(off_screen=True if image_dir else False, window_size=[1024, 600])

    # Load the example surface
    surf = aerocaps.examples.bezier_surface.bezier_surface_2x3()

    surf.plot_surface(plot, color="#4d86b8", opacity=0.5)
    surf.plot_control_points(plot, render_points_as_spheres=True, point_size=20, color="black")
    surf.plot_control_point_mesh_lines(plot, color="#d6693a")
    plot.camera.position = [-0.5, -0.7, 0.5]
    plot.camera.focal_point = [0.6, 0.5, -0.05]

    if image_dir is not None:
        plot.screenshot(os.path.join(image_dir, "bezier_surf_2x3.png"), scale=1)
    else:
        plot.show()


def plot_bezier_surf_2x3_mesh_only(image_dir: str = None):
    r"""
    Plots an :math:`(n \times m)=(2 \times 3)` Bézier surface with the isoparametric curves.

    Parameters
    ----------
    image_dir: str or None
        Where to store the output image. If no location is specified, the image file will not be saved, and the
        plot will be displayed on-screen instead. Default: ``None``
    """
    plot = pv.Plotter(off_screen=True if image_dir else False, window_size=[1024, 600])

    # Load the example surface
    surf = aerocaps.examples.bezier_surface.bezier_surface_2x3()

    surf.plot_surface(plot, color="#4d86b8", opacity=0.5, show_edges=True)
    plot.add_arrows(np.array([[0.1, 0.1, 0.05], [0.1, 0.1, 0.05]]), np.array([[0.0, 1.0, 0.0], [1.0, 0.0, 0.0]]),
                    mag=0.2, color="black")
    plot.add_point_labels(np.array([[0.1, 0.32, 0.05], [0.32, 0.1, 0.05]]), ["u", "v"], font_size=32)
    plot.camera.position = [-0.5, -0.7, 0.5]
    plot.camera.focal_point = [0.6, 0.5, -0.05]

    if image_dir is not None:
        plot.screenshot(os.path.join(image_dir, "bezier_surf_2x3_mesh_only.png"), scale=1)
    else:
        plot.show()


def main(*args, **kwargs):
    plot_bezier_surf_2x3(*args, **kwargs)
    # plot_bezier_surf_2x3_mesh_only(*args, **kwargs)


if __name__ == "__main__":
    # main()
    main(image_dir=r"C:\Users\mlaue\Documents\aerocaps\docs\source\images")



import typing
from abc import abstractmethod

import numpy as np
import pyvista as pv
from matplotlib import pyplot as plt
from scipy.optimize import fsolve

import aerocaps.iges
import aerocaps.iges.curves
import aerocaps.iges.entity
from aerocaps.geom import Geometry2D, Geometry3D, NegativeWeightError
from aerocaps.geom.point import Point2D, Point3D
from aerocaps.geom.transformation import Transformation2D, Transformation3D
from aerocaps.geom.vector import Vector3D, Vector2D
from aerocaps.units.angle import Angle
from aerocaps.units.length import Length
from aerocaps.utils.math import nchoosek


__all__ = [
    "PCurveData2D",
    "PCurveData3D",
    "PCurve2D",
    "PCurve3D",
    "Line2D",
    "Line3D",
    "CircularArc2D",
    "Bezier2D",
    "Bezier3D",
    "BSpline3D",
    "RationalBezierCurve3D",
    "NURBSCurve3D",
    "CompositeCurve3D",
    "CurveOnParametricSurface"
]


class PCurveData2D:
    def __init__(self, t: np.ndarray, xy: np.ndarray, xpyp: np.ndarray, xppypp: np.ndarray, k: np.ndarray,
                 R: np.ndarray):
        self.t = t
        self.xy = xy
        self.xpyp = xpyp
        self.xppypp = xppypp
        self.k = k
        self.R = R
        self.R_abs_min = np.min(np.abs(self.R))

    def plot(self, ax: plt.Axes, **kwargs):
        ax.plot(self.xy[:, 0], self.xy[:, 1], **kwargs)

    def get_curvature_comb(self, max_k_normalized_scale_factor, interval: int = 1):
        first_deriv_mag = np.hypot(self.xpyp[:, 0], self.xpyp[:, 1])
        comb_heads_x = self.xy[:, 0] - self.xpyp[:, 1] / first_deriv_mag * self.k * max_k_normalized_scale_factor
        comb_heads_y = self.xy[:, 1] + self.xpyp[:, 0] / first_deriv_mag * self.k * max_k_normalized_scale_factor
        # Stack the x and y columns (except for the last x and y values) horizontally and keep only the rows by the
        # specified interval:
        comb_tails = np.column_stack((self.xy[:, 0], self.xy[:, 1]))[:-1:interval, :]
        comb_heads = np.column_stack((comb_heads_x, comb_heads_y))[:-1:interval, :]
        # Add the last x and y values onto the end (to make sure they do not get skipped with input interval)
        comb_tails = np.vstack((comb_tails, np.array([self.xy[-1, 0], self.xy[-1, 1]])))
        comb_heads = np.vstack((comb_heads, np.array([comb_heads_x[-1], comb_heads_y[-1]])))
        return comb_tails, comb_heads

    def approximate_arc_length(self):
        return np.sum(np.hypot(self.xy[1:, 0] - self.xy[:-1, 0], self.xy[1:, 1] - self.xy[:-1, 1]))


class PCurve2D(Geometry2D):
    @abstractmethod
    def evaluate_point2d(self, t: float) -> Point2D:
        pass

    @abstractmethod
    def evaluate_single_t(self, t: float) -> np.ndarray:
        pass

    @abstractmethod
    def evaluate(self, t: np.ndarray = None) -> PCurveData2D:
        pass


class PCurveData3D:
    def __init__(self, t: np.ndarray, xyz: np.ndarray, xpypzp: np.ndarray, xppyppzpp: np.ndarray, k: np.ndarray,
                 R: np.ndarray):
        self.t = t
        self.xyz = xyz
        self.xpypzp = xpypzp
        self.xppyppzpp = xppyppzpp
        self.k = k
        self.R = R
        self.R_abs_min = np.min(np.abs(self.R))

    def plot(self, ax: plt.Axes, **kwargs):
        ax.plot3D(self.xyz[:, 0], self.xyz[:, 1], self.xyz[:, 2], **kwargs)

    def get_curvature_comb(self, max_k_normalized_scale_factor, interval: int = 1):
        first_deriv_mag = np.sqrt(self.xpypzp[:, 0] ** 2, self.xpypzp[:, 1] ** 2, self.xpypzp[:, 2] ** 2)
        comb_heads_x = self.xyz[:, 0] - self.xpypzp[:, 1] / first_deriv_mag * self.k * max_k_normalized_scale_factor
        comb_heads_y = self.xyz[:, 1] + self.xpypzp[:, 0] / first_deriv_mag * self.k * max_k_normalized_scale_factor
        comb_heads_z = self.xyz[:, 2] + self.xpypzp[:, 0] / first_deriv_mag * self.k * max_k_normalized_scale_factor
        # Stack the x and y columns (except for the last x and y values) horizontally and keep only the rows by the
        # specified interval:
        comb_tails = np.column_stack((self.xyz[:, 0], self.xyz[:, 1], self.xyz[:, 2]))[:-1:interval, :]
        comb_heads = np.column_stack((comb_heads_x, comb_heads_y))[:-1:interval, :]
        # Add the last x and y values onto the end (to make sure they do not get skipped with input interval)
        comb_tails = np.vstack((comb_tails, np.array([self.xyz[-1, 0], self.xyz[-1, 1], self.xyz[-1, 2]])))
        comb_heads = np.vstack((comb_heads, np.array([comb_heads_x[-1], comb_heads_y[-1], comb_heads_z[-1]])))
        return comb_tails, comb_heads

    def approximate_arc_length(self):
        return np.sum(np.sqrt(
            (self.xyz[1:, 0] - self.xyz[:-1, 0]) ** 2,
            (self.xyz[1:, 1] - self.xyz[:-1, 1]) ** 2,
            (self.xyz[1:, 2] - self.xyz[:-1, 2]) ** 2
        ))


class PCurve3D(Geometry3D):
    @abstractmethod
    def evaluate_point3d(self, t: float) -> Point3D:
        pass

    @abstractmethod
    def evaluate_single_t(self, t: float) -> np.ndarray:
        pass

    @abstractmethod
    def evaluate(self, t: np.ndarray = None) -> PCurveData3D:
        pass


class Line2D(PCurve2D):
    def __init__(self,
                 p0: Point2D,
                 theta: Angle = None,
                 p1: Point2D = None,
                 d: Length = Length(m=1.0)
                 ):
        if theta and p1:
            raise ValueError("Angle theta should not be specified if p1 is specified")
        if not theta and not p1:
            raise ValueError("Must specify either angle theta or p1")
        self.p0 = p0
        self.theta = theta
        from aerocaps.geom.tools import measure_distance_between_points  # Avoid circular import
        self.d = d if not p1 else Length(m=measure_distance_between_points(p0, p1))
        self.p1 = self.evaluate_point2d(1.0) if not p1 else p1
        self.control_points = [self.p0, self.p1]

    def evaluate_point2d(self, t: float) -> Point2D:
        if self.theta:
            return Point2D(
                x=self.p0.x + self.d * np.cos(self.theta.rad) * t,
                y=self.p0.y + self.d * np.sin(self.theta.rad) * t
            )
        else:
            return self.p0 + t * (self.p1 - self.p0)

    def evaluate_single_t(self, t: float) -> np.ndarray:
        return self.evaluate_point2d(t).as_array()

    def evaluate(self, t: np.ndarray = None) -> PCurveData2D:
        t = np.linspace(0.0, 1.0, 10) if t is None else t

        if self.theta:
            x = self.p0.x.m + self.d.m * np.cos(self.theta.rad) * t
            y = self.p0.y.m + self.d.m * np.sin(self.theta.rad) * t
            xp = self.d.m * np.cos(self.theta.rad) * np.ones(t.shape)
            yp = self.d.m * np.sin(self.theta.rad) * np.ones(t.shape)
        else:
            x = self.p0.x.m + t * (self.p1.x.m - self.p0.x.m)
            y = self.p0.y.m + t * (self.p1.y.m - self.p0.y.m)
            xp = (self.p1.x.m - self.p0.x.m) * np.ones(t.shape)
            yp = (self.p1.y.m - self.p0.y.m) * np.ones(t.shape)

        xy = np.column_stack((x, y))
        xpyp = np.column_stack((xp, yp))
        xppypp = np.zeros((len(t), 2))
        R = np.inf * np.ones(t.shape)
        k = np.zeros(t.shape)
        return PCurveData2D(t=t, xy=xy, xpyp=xpyp, xppypp=xppypp, k=k, R=R)

    def get_vector(self) -> Vector2D:
        return Vector2D(p0=self.p0, p1=self.p1)

    def plot(self, plot: pv.Plotter, **line_kwargs):
        line_arr = np.array([self.p0.as_array(), self.p1.as_array()])
        plot.add_lines(line_arr, **line_kwargs)


class Line3D(PCurve3D):
    def __init__(self,
                 p0: Point3D,
                 theta: Angle = None,
                 phi: Angle = None,
                 p1: Point3D = None,
                 d: Length = Length(m=1.0)
                 ):
        if (theta and p1) or (phi and p1):
            raise ValueError("Angles should not be specified if p1 is specified")
        if (not theta and not p1) or (not phi and not p1):
            raise ValueError("Must specify either both angles, theta and phi, or p1")
        self.p0 = p0
        self.theta = theta
        self.phi = phi
        from aerocaps.geom.tools import measure_distance_between_points  # Avoid circular import
        self.d = d if not p1 else Length(m=measure_distance_between_points(p0, p1))
        self.p1 = self.evaluate_point3d(1.0) if not p1 else p1
        self.control_points = [self.p0, self.p1]

    def to_iges(self, *args, **kwargs) -> aerocaps.iges.entity.IGESEntity:
        return aerocaps.iges.curves.LineIGES(self.p0.as_array(), self.p1.as_array())

    def from_iges(self):
        pass

    def get_control_point_array(self, unit: str = "m") -> np.ndarray:
        return np.array([p.as_array(unit=unit) for p in self.control_points])

    def reverse(self) -> "Line3D":
        return self.__class__(p0=self.p1, p1=self.p0)

    def projection_on_principal_plane(self, plane: str = "XY") -> Line2D:
        return Line2D(p0=self.p0.projection_on_principal_plane(plane), p1=self.p1.projection_on_principal_plane(plane))

    def evaluate_point3d(self, t: float) -> Point3D:
        if self.phi and self.theta:
            return Point3D(
                x=self.p0.x + self.d * np.cos(self.phi.rad) * np.cos(self.theta.rad) * t,
                y=self.p0.y + self.d * np.cos(self.phi.rad) * np.sin(self.theta.rad) * t,
                z=self.p0.z + self.d * np.sin(self.phi.rad) * t
            )
        else:
            return self.p0 + t * (self.p1 - self.p0)

    def evaluate_single_t(self, t: float) -> np.ndarray:
        return self.evaluate_point3d(t).as_array()

    def evaluate(self, t: np.ndarray = None) -> PCurveData3D:
        t = np.linspace(0.0, 1.0, 10) if t is None else t

        if self.theta:
            x = self.p0.x + self.d * np.cos(self.phi.rad) * np.cos(self.theta.rad) * t,
            y = self.p0.y + self.d * np.cos(self.phi.rad) * np.sin(self.theta.rad) * t,
            z = self.p0.z + self.d * np.sin(self.phi.rad) * t
            xp = self.d.m * np.cos(self.phi.rad) * np.cos(self.theta.rad) * np.ones(t.shape)
            yp = self.d.m * np.cos(self.phi.rad) * np.sin(self.theta.rad) * np.ones(t.shape)
            zp = self.d.m * np.sin(self.phi.rad) * np.ones(t.shape)
        else:
            x = self.p0.x.m + t * (self.p1.x.m - self.p0.x.m)
            y = self.p0.y.m + t * (self.p1.y.m - self.p0.y.m)
            z = self.p0.z.m + t * (self.p1.z.m - self.p0.z.m)
            xp = (self.p1.x.m - self.p0.x.m) * np.ones(t.shape)
            yp = (self.p1.y.m - self.p0.y.m) * np.ones(t.shape)
            zp = (self.p1.z.m - self.p0.z.m) * np.ones(t.shape)

        xyz = np.column_stack((x, y, z))
        xpypzp = np.column_stack((xp, yp, zp))
        xppyppzpp = np.zeros((len(t), 3))
        R = np.inf * np.ones(t.shape)
        k = np.zeros(t.shape)
        return PCurveData3D(t=t, xyz=xyz, xpypzp=xpypzp, xppyppzpp=xppyppzpp, k=k, R=R)

    def get_vector(self) -> Vector3D:
        return Vector3D(p0=self.p0, p1=self.p1)

    def plot(self, plot: pv.Plotter, **line_kwargs):
        line_arr = np.array([self.p0.as_array(), self.p1.as_array()])
        plot.add_lines(line_arr, **line_kwargs)


class CircularArc2D(PCurve2D):
    def __init__(self, center: Point2D, radius: Length, start_point: Point2D = None, end_point: Point2D = None,
                 start_angle: Angle = None, end_angle: Angle = None, complement: bool = False):
        self.center = center
        self.radius = radius
        self.start_point = start_point
        self.end_point = end_point
        if start_angle is None:
            self.start_angle = Angle(rad=np.arctan2(start_point.y.m - center.y.m, start_point.x.m - center.x.m))
        else:
            self.start_angle = start_angle
        if end_angle is None:
            self.end_angle = Angle(rad=np.arctan2(end_point.y.m - center.y.m, end_point.x.m - center.x.m))
        else:
            self.end_angle = end_angle
        self.complement = complement

    def _map_t_to_angle(self, t):
        if self.complement:
            return self.start_angle.rad - (2 * np.pi - (self.end_angle.rad - self.start_angle.rad)) * t
        else:
            return (self.end_angle.rad - self.start_angle.rad) * t + self.start_angle.rad

    def evaluate_point2d(self, t: float) -> Point2D:
        return Point2D(
            x=self.center.x + self.radius * np.cos(self._map_t_to_angle(t)),
            y=self.center.y + self.radius * np.sin(self._map_t_to_angle(t)),
        )

    def evaluate_single_t(self, t: float) -> np.ndarray:
        return self.evaluate_point2d(t).as_array()

    def evaluate(self, t: np.ndarray = None) -> PCurveData2D:
        t = np.linspace(0.0, 1.0, 100) if t is None else t
        angle_range = self.end_angle.rad - self.start_angle.rad
        x = self.center.x.m + self.radius.m * np.cos(self._map_t_to_angle(t))
        y = self.center.y.m + self.radius.m * np.sin(self._map_t_to_angle(t))
        xp = -self.radius.m * angle_range * np.sin(self._map_t_to_angle(t))
        yp = self.radius.m * angle_range * np.cos(self._map_t_to_angle(t))
        xpp = -self.radius.m * angle_range ** 2 * np.cos(self._map_t_to_angle(t))
        ypp = -self.radius.m * angle_range ** 2 * np.sin(self._map_t_to_angle(t))
        xy = np.column_stack((x, y))
        xpyp = np.column_stack((xp, yp))
        xppypp = np.column_stack((xpp, ypp))
        R = self.radius.m * np.ones(t.shape)
        k = 1 / self.radius.m * np.ones(t.shape)
        return PCurveData2D(t=t, xy=xy, xpyp=xpyp, xppypp=xppypp, k=k, R=R)


class Bezier2D(PCurve2D):

    def __init__(self, control_points: typing.List[Point2D]):
        self.control_points = control_points
        self.curve_connections = []

    @property
    def degree(self):
        return len(self.control_points) - 1

    @degree.setter
    def degree(self, value):
        raise AttributeError("The 'degree' property is read-only. Use the Bezier2D.elevate_degree method to increase"
                             "the degree of the curve while retaining the shape, or manually add or remove control "
                             "points to change the degree directly.")

    @staticmethod
    def bernstein_poly(n: int, i: int, t: int or float or np.ndarray):
        r"""
        Calculates the Bernstein polynomial for a given Bézier curve order, index, and parameter vector. The
        Bernstein polynomial is described by

        .. math::

            B_{i,n}(t)={n \choose i} t^i (1-t)^{n-i}

        Arguments
        =========
        n: int
            Bézier curve degree (one less than the number of control points in the Bézier curve)
        i: int
            Bézier curve index
        t: int, float, or np.ndarray
            Parameter vector for the Bézier curve

        Returns
        =======
        np.ndarray
            Array of values of the Bernstein polynomial evaluated for each point in the parameter vector
        """
        return nchoosek(n, i) * t ** i * (1.0 - t) ** (n - i)

    @staticmethod
    def finite_diff_P(P: np.ndarray, k: int, i: int):
        """Calculates the finite difference of the control points as shown in
        https://pages.mtu.edu/~shene/COURSES/cs3621/NOTES/spline/Bezier/bezier-der.html

        Arguments
        =========
        P: np.ndarray
            Array of control points for the Bézier curve
        k: int
            Finite difference level (e.g., k = 1 is the first derivative finite difference)
        i: int
            An index referencing a location in the control point array
        """

        def finite_diff_recursive(_k, _i):
            if _k > 1:
                return finite_diff_recursive(_k - 1, _i + 1) - finite_diff_recursive(_k - 1, _i)
            else:
                return P[_i + 1, :] - P[_i, :]

        return finite_diff_recursive(k, i)

    def derivative(self, P: np.ndarray, t: np.ndarray, degree: int, order: int):
        r"""
        Calculates an arbitrary-order derivative of the Bézier curve

        Parameters
        ==========
        P: np.ndarray
            The control point array

        t: np.ndarray
            The parameter vector

        degree: int
            The degree of the Bézier curve

        order: int
            The derivative order. For example, ``order=2`` returns the second derivative.

        Returns
        =======
        np.ndarray
            An array of ``shape=(N,2)`` where ``N`` is the number of evaluated points specified by the :math:`t` vector.
            The columns represent :math:`C^{(m)}_x(t)` and :math:`C^{(m)}_y(t)`, where :math:`m` is the
            derivative order.
        """
        return np.sum(np.array([np.prod(np.array([degree - idx for idx in range(order)])) *
                                np.array([self.finite_diff_P(P, order, i)]).T *
                                np.array([self.bernstein_poly(degree - order, i, t)])
                                for i in range(degree + 1 - order)]), axis=0).T

    def get_control_point_array(self, unit: str = "m") -> np.ndarray:
        return np.array([p.as_array(unit=unit) for p in self.control_points])

    @classmethod
    def generate_from_array(cls, P: np.ndarray, unit: str = "m"):
        return cls([Point2D(x=Length(**{unit: xy[0]}), y=Length(**{unit: xy[1]})) for xy in P])

    def evaluate_point2d(self, t: float) -> Point2D:
        n_ctrl_points = len(self.control_points)
        degree = n_ctrl_points - 1
        P = self.get_control_point_array()

        # Evaluate the curve
        x, y = 0.0, 0.0
        for i in range(n_ctrl_points):
            # Calculate the x- and y-coordinates of the Bézier curve given the input vector t
            x += P[i, 0] * self.bernstein_poly(degree, i, t)
            y += P[i, 1] * self.bernstein_poly(degree, i, t)

        return Point2D(x=Length(m=x), y=Length(m=y))

    def evaluate_single_t(self, t: float) -> np.ndarray:
        return self.evaluate_point2d(t).as_array()

    def evaluate(self, t: np.array or None = None) -> PCurveData2D:
        r"""
        Evaluates the curve using an optionally specified parameter vector.

        Parameters
        ==========
        t: np.ndarray or ``None``
            Optional direct specification of the parameter vector for the curve. Not specifying this value
            gives a linearly spaced parameter vector from ``t_start`` or ``t_end`` with the default size.
            Default: ``None``

        Returns
        =======
        PCurveData
            Data class specifying the following information about the Bézier curve:

            .. math::

                    C_x(t), C_y(t), C'_x(t), C'_y(t), C''_x(t), C''_y(t), \kappa(t)

            where the :math:`x` and :math:`y` subscripts represent the :math:`x` and :math:`y` components of the
            vector-valued functions :math:`\vec{C}(t)`, :math:`\vec{C}'(t)`, and :math:`\vec{C}''(t)`.
        """
        # Pass the starting and ending parameter vector values to the parameter vector generator if they were
        # specified directly

        # Generate the parameter vector
        t = np.linspace(0.0, 1.0, 100) if t is None else t

        # Number of control points, curve degree, control point array
        n_ctrl_points = len(self.control_points)
        degree = n_ctrl_points - 1
        P = self.get_control_point_array()

        # Evaluate the curve
        x, y = np.zeros(t.shape), np.zeros(t.shape)
        for i in range(n_ctrl_points):
            # Calculate the x- and y-coordinates of the Bézier curve given the input vector t
            x += P[i, 0] * self.bernstein_poly(degree, i, t)
            y += P[i, 1] * self.bernstein_poly(degree, i, t)
        xy = np.column_stack((x, y))

        # Calculate the first derivative
        first_deriv = self.derivative(P=P, t=t, degree=degree, order=1)
        xp = first_deriv[:, 0]
        yp = first_deriv[:, 1]

        # Calculate the second derivative
        second_deriv = self.derivative(P=P, t=t, degree=degree, order=2)
        xpp = second_deriv[:, 0]
        ypp = second_deriv[:, 1]

        # Combine the derivative x and y data
        xpyp = np.column_stack((xp, yp))
        xppypp = np.column_stack((xpp, ypp))

        # Calculate the curvature
        with np.errstate(divide='ignore', invalid='ignore'):
            # Calculate the curvature of the Bézier curve (k = kappa = 1 / R, where R is the radius of curvature)
            k = np.true_divide((xp * ypp - yp * xpp), (xp ** 2 + yp ** 2) ** (3 / 2))

        # Calculate the radius of curvature: R = 1 / kappa
        with np.errstate(divide='ignore', invalid='ignore'):
            R = np.true_divide(1, k)

        return PCurveData2D(t=t, xy=xy, xpyp=xpyp, xppypp=xppypp, k=k, R=R)

    def compute_t_corresponding_to_x(self, x_seek: float, t0: float = 0.5):
        def bez_root_find_func(t):
            point = self.evaluate_point2d(t[0])
            return np.array([point.x.m - x_seek])

        return fsolve(bez_root_find_func, x0=np.array([t0]))[0]

    def compute_t_corresponding_to_y(self, y_seek: float, t0: float = 0.5):
        def bez_root_find_func(t):
            point = self.evaluate_point2d(t[0])
            return np.array([point.y.m - y_seek])

        return fsolve(bez_root_find_func, x0=np.array([t0]))[0]

    def convert_to_3d(self, plane: str = "XY"):
        valid_planes = ["XY", "YZ", "XZ"]
        plane_axis_mapping = {k: v for k, v in zip(valid_planes, [2, 0, 1])}
        if plane not in valid_planes:
            raise ValueError(f"Plane must be one of {valid_planes}. Given plane was {plane}")
        P = self.get_control_point_array()
        new_P = np.insert(P, plane_axis_mapping[plane], 0.0, axis=1)
        return Bezier3D.generate_from_array(new_P)

    def transform(self, **transformation_kwargs):
        transformation = Transformation2D(**transformation_kwargs)
        return Bezier2D.generate_from_array(transformation.transform(self.get_control_point_array()))

    def elevate_degree(self) -> "Bezier2D":
        """
        Elevates the degree of the Bézier curve. See algorithm source
        `here <https://pages.mtu.edu/~shene/COURSES/cs3621/NOTES/spline/Bezier/bezier-elev.html>`_.

        Returns
        -------
        Bezier2D
            A new Bézier curve with identical shape to the current one but with one additional control point.
        """
        n = self.degree
        P = self.get_control_point_array()

        # New array has one additional control point (current array only has n+1 control points)
        new_control_points = np.zeros((P.shape[0] + 1, P.shape[1]))

        # Set starting and ending control points to what they already were
        new_control_points[0, :] = P[0, :]
        new_control_points[-1, :] = P[-1, :]

        # Update all the other control points
        for i in range(1, n + 1):  # 1 <= i <= n
            new_control_points[i, :] = i / (n + 1) * P[i - 1, :] + (1 - i / (n + 1)) * P[i, :]

        return Bezier2D.generate_from_array(new_control_points)

    def split(self, t_split: float):

        # Number of control points, curve degree, control point array
        n_ctrl_points = len(self.control_points)
        degree = n_ctrl_points - 1
        P = np.array([p.as_array() for p in self.control_points])

        def de_casteljau(i: int, j: int) -> np.ndarray:
            """
            Based on https://en.wikipedia.org/wiki/De_Casteljau%27s_algorithm. Recursive algorithm where the
            base case is just the value of the ith original control point.

            Parameters
            ----------
            i: int
                Lower index
            j: int
                Upper index

            Returns
            -------
            np.ndarray
                A one-dimensional array containing the :math:`x` and :math:`y` values of a control point evaluated
                at :math:`(i,j)` for a Bézier curve split at the parameter value ``t_split``
            """
            if j == 0:
                return P[i, :]
            return de_casteljau(i, j - 1) * (1 - t_split) + de_casteljau(i + 1, j - 1) * t_split

        bez_split_1_P = np.array([de_casteljau(i=0, j=i) for i in range(n_ctrl_points)])
        bez_split_2_P = np.array([de_casteljau(i=i, j=degree - i) for i in range(n_ctrl_points)])

        bez_1_points = [self.control_points[0]] + [Point2D(Length(m=xy[0]), Length(m=xy[1])) for xy in bez_split_1_P[1:, :]]
        bez_2_points = [bez_1_points[-1]] + [Point2D(Length(m=xy[0]), Length(m=xy[1])) for xy in bez_split_2_P[1:-1, :]] + [
            self.control_points[-1]]

        return (
            Bezier2D(bez_1_points),
            Bezier2D(bez_2_points)
        )


class Bezier3D(PCurve3D):

    def __init__(self, control_points: typing.List[Point3D]):
        self.control_points = control_points
        self.curve_connections = []

    @property
    def degree(self):
        return len(self.control_points) - 1

    @degree.setter
    def degree(self, value):
        raise AttributeError("The 'degree' property is read-only. Use the Bezier3D.elevate_degree method to increase"
                             "the degree of the curve while retaining the shape, or manually add or remove control "
                             "points to change the degree directly.")

    def to_iges(self, *args, **kwargs) -> aerocaps.iges.entity.IGESEntity:
        return aerocaps.iges.curves.BezierIGES(
            control_points_XYZ=self.get_control_point_array(),
        )

    def reverse(self) -> "Bezier3D":
        return self.__class__(self.control_points[::-1])

    def projection_on_principal_plane(self, plane: str = "XY") -> Bezier2D:
        return Bezier2D(control_points=[pt.projection_on_principal_plane(plane) for pt in self.control_points])

    @staticmethod
    def bernstein_poly(n: int, i: int, t: int or float or np.ndarray):
        r"""
        Calculates the Bernstein polynomial for a given Bézier curve order, index, and parameter vector. The
        Bernstein polynomial is described by

        .. math::

            B_{i,n}(t)={n \choose i} t^i (1-t)^{n-i}

        Arguments
        =========
        n: int
            Bézier curve degree (one less than the number of control points in the Bézier curve)
        i: int
            Bézier curve index
        t: int, float, or np.ndarray
            Parameter vector for the Bézier curve

        Returns
        =======
        np.ndarray
            Array of values of the Bernstein polynomial evaluated for each point in the parameter vector
        """
        return nchoosek(n, i) * t ** i * (1.0 - t) ** (n - i)

    @staticmethod
    def finite_diff_P(P: np.ndarray, k: int, i: int):
        """Calculates the finite difference of the control points as shown in
        https://pages.mtu.edu/~shene/COURSES/cs3621/NOTES/spline/Bezier/bezier-der.html

        Arguments
        =========
        P: np.ndarray
            Array of control points for the Bézier curve
        k: int
            Finite difference level (e.g., k = 1 is the first derivative finite difference)
        i: int
            An index referencing a location in the control point array
        """

        def finite_diff_recursive(_k, _i):
            if _k > 1:
                return finite_diff_recursive(_k - 1, _i + 1) - finite_diff_recursive(_k - 1, _i)
            else:
                return P[_i + 1, :] - P[_i, :]

        return finite_diff_recursive(k, i)

    def derivative(self, P: np.ndarray, t: np.ndarray, degree: int, order: int):
        r"""
        Calculates an arbitrary-order derivative of the Bézier curve

        Parameters
        ==========
        P: np.ndarray
            The control point array

        t: np.ndarray
            The parameter vector

        degree: int
            The degree of the Bézier curve

        order: int
            The derivative order. For example, ``order=2`` returns the second derivative.

        Returns
        =======
        np.ndarray
            An array of ``shape=(N,2)`` where ``N`` is the number of evaluated points specified by the :math:`t` vector.
            The columns represent :math:`C^{(m)}_x(t)` and :math:`C^{(m)}_y(t)`, where :math:`m` is the
            derivative order.
        """
        return np.sum(np.array([np.prod(np.array([degree - idx for idx in range(order)])) *
                                np.array([self.finite_diff_P(P, order, i)]).T *
                                np.array([self.bernstein_poly(degree - order, i, t)])
                                for i in range(degree + 1 - order)]), axis=0).T

    def get_control_point_array(self, unit: str = "m") -> np.ndarray:
        return np.array([p.as_array(unit=unit) for p in self.control_points])

    @classmethod
    def generate_from_array(cls, P: np.ndarray, unit: str = "m"):
        return cls([Point3D(x=Length(**{unit: xyz[0]}),
                            y=Length(**{unit: xyz[1]}),
                            z=Length(**{unit: xyz[2]})) for xyz in P])

    def evaluate_point3d(self, t: float) -> Point3D:
        n_ctrl_points = len(self.control_points)
        degree = n_ctrl_points - 1
        P = self.get_control_point_array()

        # Evaluate the curve
        x, y, z = 0.0, 0.0, 0.0
        for i in range(n_ctrl_points):
            # Calculate the x- and y-coordinates of the Bézier curve given the input vector t
            x += P[i, 0] * self.bernstein_poly(degree, i, t)
            y += P[i, 1] * self.bernstein_poly(degree, i, t)
            z += P[i, 2] * self.bernstein_poly(degree, i, t)

        return Point3D(x=Length(m=x), y=Length(m=y), z=Length(m=z))

    def evaluate_single_t(self, t: float) -> np.ndarray:
        return self.evaluate_point3d(t).as_array()

    def evaluate(self, t: np.array = None) -> PCurveData3D:
        r"""
        Evaluates the curve using an optionally specified parameter vector.

        Parameters
        ==========
        t: np.ndarray or ``None``
            Optional direct specification of the parameter vector for the curve. Not specifying this value
            gives a linearly spaced parameter vector from ``t_start`` or ``t_end`` with the default size.
            Default: ``None``

        Returns
        =======
        PCurveData
            Data class specifying the following information about the Bézier curve:

            .. math::

                    C_x(t), C_y(t), C'_x(t), C'_y(t), C''_x(t), C''_y(t), \kappa(t)

            where the :math:`x` and :math:`y` subscripts represent the :math:`x` and :math:`y` components of the
            vector-valued functions :math:`\vec{C}(t)`, :math:`\vec{C}'(t)`, and :math:`\vec{C}''(t)`.
        """
        # Pass the starting and ending parameter vector values to the parameter vector generator if they were
        # specified directly

        # Generate the parameter vector
        t = np.linspace(0.0, 1.0, 100) if t is None else t

        # Number of control points, curve degree, control point array
        n_ctrl_points = len(self.control_points)
        degree = n_ctrl_points - 1
        P = self.get_control_point_array()

        # Evaluate the curve
        x, y, z = np.zeros(t.shape), np.zeros(t.shape), np.zeros(t.shape)
        for i in range(n_ctrl_points):
            # Calculate the x- and y-coordinates of the Bézier curve given the input vector t
            x += P[i, 0] * self.bernstein_poly(degree, i, t)
            y += P[i, 1] * self.bernstein_poly(degree, i, t)
            z += P[i, 2] * self.bernstein_poly(degree, i, t)
        xyz = np.column_stack((x, y, z))

        # Calculate the first derivative
        first_deriv = self.derivative(P=P, t=t, degree=degree, order=1)
        xp = first_deriv[:, 0]
        yp = first_deriv[:, 1]
        zp = first_deriv[:, 2]

        # Calculate the second derivative
        second_deriv = self.derivative(P=P, t=t, degree=degree, order=2)
        xpp = second_deriv[:, 0]
        ypp = second_deriv[:, 1]
        zpp = second_deriv[:, 2]

        # Combine the derivative x and y data
        xpypzp = np.column_stack((xp, yp, zp))
        xppyppzpp = np.column_stack((xpp, ypp, zpp))

        # Calculate the curvature
        with np.errstate(divide='ignore', invalid='ignore'):
            # Calculate the curvature of the Bézier curve (k = kappa = 1 / R, where R is the radius of curvature)
            k = np.true_divide(
                np.linalg.norm(np.cross(xpypzp, xppyppzpp), axis=1),
                np.linalg.norm(xpypzp, axis=1) ** 3
            )

        # Calculate the radius of curvature: R = 1 / kappa
        with np.errstate(divide='ignore', invalid='ignore'):
            R = np.true_divide(1, k)

        return PCurveData3D(t=t, xyz=xyz, xpypzp=xpypzp, xppyppzpp=xppyppzpp, k=k, R=R)

    def compute_t_corresponding_to_x(self, x_seek: float, t0: float = 0.5):
        def bez_root_find_func(t):
            point = self.evaluate_point3d(t[0])
            return np.array([point.x.m - x_seek])

        return fsolve(bez_root_find_func, x0=np.array([t0]))[0]

    def compute_t_corresponding_to_y(self, y_seek: float, t0: float = 0.5):
        def bez_root_find_func(t):
            point = self.evaluate_point3d(t[0])
            return np.array([point.y.m - y_seek])

        return fsolve(bez_root_find_func, x0=np.array([t0]))[0]

    def compute_t_corresponding_to_z(self, z_seek: float, t0: float = 0.5):
        def bez_root_find_func(t):
            point = self.evaluate_point3d(t[0])
            return np.array([point.z.m - z_seek])

        return fsolve(bez_root_find_func, x0=np.array([t0]))[0]

    def transform(self, **transformation_kwargs) -> "Bezier3D":
        transformation = Transformation3D(**transformation_kwargs)
        return Bezier3D.generate_from_array(transformation.transform(self.get_control_point_array()))

    def elevate_degree(self) -> "Bezier3D":
        """
        Elevates the degree of the Bézier curve. See algorithm source
        `here <https://pages.mtu.edu/~shene/COURSES/cs3621/NOTES/spline/Bezier/bezier-elev.html>`_.

        Returns
        -------
        Bezier3D
            A new Bézier curve with identical shape to the current one but with one additional control point.
        """
        n = self.degree
        P = self.get_control_point_array()

        # New array has one additional control point (current array only has n+1 control points)
        new_control_points = np.zeros((P.shape[0] + 1, P.shape[1]))

        # Set starting and ending control points to what they already were
        new_control_points[0, :] = P[0, :]
        new_control_points[-1, :] = P[-1, :]

        # Update all the other control points
        for i in range(1, n + 1):  # 1 <= i <= n
            new_control_points[i, :] = i / (n + 1) * P[i - 1, :] + (1 - i / (n + 1)) * P[i, :]

        return Bezier3D.generate_from_array(new_control_points)

    def split(self, t_split: float) -> ("Bezier3D", "Bezier3D"):

        # Number of control points, curve degree, control point array
        n_ctrl_points = len(self.control_points)
        degree = n_ctrl_points - 1
        P = np.array([p.as_array() for p in self.control_points])

        def de_casteljau(i: int, j: int) -> np.ndarray:
            """
            Based on https://en.wikipedia.org/wiki/De_Casteljau%27s_algorithm. Recursive algorithm where the
            base case is just the value of the ith original control point.

            Parameters
            ----------
            i: int
                Lower index
            j: int
                Upper index

            Returns
            -------
            np.ndarray
                A one-dimensional array containing the :math:`x` and :math:`y` values of a control point evaluated
                at :math:`(i,j)` for a Bézier curve split at the parameter value ``t_split``
            """
            if j == 0:
                return P[i, :]
            return de_casteljau(i, j - 1) * (1 - t_split) + de_casteljau(i + 1, j - 1) * t_split

        bez_split_1_P = np.array([de_casteljau(i=0, j=i) for i in range(n_ctrl_points)])
        bez_split_2_P = np.array([de_casteljau(i=i, j=degree - i) for i in range(n_ctrl_points)])

        bez_1_points = [self.control_points[0]] + [Point3D(
            Length(m=xyz[0]), Length(m=xyz[1]), Length(m=xyz[2])) for xyz in bez_split_1_P[1:, :]]
        bez_2_points = [bez_1_points[-1]] + [Point3D(
            Length(m=xyz[0]), Length(m=xyz[1]), Length(m=xyz[2])) for xyz in bez_split_2_P[1:-1, :]] + [self.control_points[-1]]

        return (
            Bezier3D(bez_1_points),
            Bezier3D(bez_2_points)
        )


class NURBSCurve3D(Geometry3D):
    def __init__(self,
                 control_points: np.ndarray,
                 weights: np.ndarray,
                 knot_vector: np.ndarray,
                 degree: int):
        """
        Non-uniform rational B-spline (NURBS) curve evaluation class
        """
        assert control_points.ndim == 2
        assert weights.ndim == 1
        assert knot_vector.ndim == 1
        assert len(knot_vector) == len(control_points) + degree + 1
        assert len(control_points) == len(weights)

        # Negative weight check
        for weight in weights:
            if weight < 0:
                raise NegativeWeightError("All weights must be non-negative")

        self.control_points = control_points
        self.dim = self.control_points.shape[1]
        self.weights = np.array(weights)
        self.knot_vector = np.array(knot_vector)
        self.degree = degree
        self.possible_spans, self.possible_span_indices = self._get_possible_spans()

    def to_iges(self, *args, **kwargs) -> aerocaps.iges.entity.IGESEntity:
        return aerocaps.iges.curves.RationalBSplineCurveIGES(
            knots=self.knot_vector,
            weights=self.weights,
            control_points_XYZ=self.control_points,
            degree=self.degree
        )

    def reverse(self) -> "NURBSCurve3D":
        return self.__class__(np.flipud(self.control_points),
                              self.weights[::-1],
                              (1.0 - self.knot_vector)[::-1],
                              self.degree)

    def evaluate_ndarray(self, t: float) -> np.ndarray:
        """
        Evaluate the NURBS curve at parameter t
        """
        B = self._basis_functions(t, self.degree)
        point = np.dot(B * self.weights, self.control_points) / np.sum(B * self.weights)
        return point

    def evaluate_simple(self, t: float) -> Point3D:
        """
        Evaluates the NURBS curve at parameter t as a ``Point3D``

        Parameters
        ----------
        t: float
            Parameter value

        Returns
        -------
        Point3D
            Value of the NURBS curve at the specified parameter value
        """
        return Point3D.from_array(self.evaluate_ndarray(t))

    def evaluate(self, t_vec: np.ndarray) -> np.ndarray:
        """
        Evaluates the NURBS curve at a vector of parameter values
        """
        points = np.array([self.evaluate_ndarray(t) for t in t_vec])
        return points

    def _get_possible_spans(self) -> (np.ndarray, np.ndarray):
        possible_span_indices = np.array([], dtype=int)
        possible_spans = []
        for knot_idx, (knot_1, knot_2) in enumerate(zip(self.knot_vector[:-1], self.knot_vector[1:])):
            if knot_1 == knot_2:
                continue
            possible_span_indices = np.append(possible_span_indices, knot_idx)
            possible_spans.append([knot_1, knot_2])
        return np.array(possible_spans), possible_span_indices

    def _cox_de_boor(self, t: float, i: int, p: int) -> float:
        if p == 0:
            return 1.0 if i in self.possible_span_indices and self._find_span(t) == i else 0.0
        else:
            with np.errstate(divide="ignore", invalid="ignore"):
                f = (t - self.knot_vector[i]) / (self.knot_vector[i + p] - self.knot_vector[i])
                g = (self.knot_vector[i + p + 1] - t) / (self.knot_vector[i + p + 1] - self.knot_vector[i + 1])
                if np.isinf(f) or np.isnan(f):
                    f = 0.0
                if np.isinf(g) or np.isnan(g):
                    g = 0.0
                if f == 0.0 and g == 0.0:
                    return 0.0
                elif f != 0.0 and g == 0.0:
                    return f * self._cox_de_boor(t, i, p - 1)
                elif f == 0.0 and g != 0.0:
                    return g * self._cox_de_boor(t, i + 1, p - 1)
                else:
                    return f * self._cox_de_boor(t, i, p - 1) + g * self._cox_de_boor(t, i + 1, p - 1)

    def _basis_functions(self, t: float, p: int):
        """
        Compute the non-zero basis functions at parameter t
        """
        return np.array([self._cox_de_boor(t, i, p) for i in range(self.control_points.shape[0])])

    def _find_span(self, t: float):
        """
        Find the knot span index
        """
        # insertion_point = bisect.bisect_left(self.non_repeated_knots, t)
        # return self.possible_spans[insertion_point - 1]
        for knot_span, knot_span_idx in zip(self.possible_spans, self.possible_span_indices):
            if knot_span[0] <= t < knot_span[1]:
                return knot_span_idx
        if t == self.possible_spans[-1][1]:
            return self.possible_span_indices[-1]


class RationalBezierCurve3D(Geometry3D):

    projection_dict = {
        "X": 0,
        "Y": 1,
        "Z": 2,
    }

    def __init__(self,
                 control_points: typing.List[Point3D],
                 weights: np.ndarray):
        """
        Non-uniform rational B-spline (NURBS) curve evaluation class
        """
        self.control_points = control_points
        assert weights.ndim == 1
        assert len(control_points) == len(weights)

        # Negative weight check
        for weight in weights:
            if weight < 0:
                raise NegativeWeightError("All weights must be non-negative")

        self.dim = 3
        self.weights = np.array(weights)
        self.knot_vector = np.zeros(2 * len(control_points))
        self.knot_vector[len(control_points):] = 1.0
        self.degree = len(control_points) - 1
        assert self.knot_vector.ndim == 1
        assert len(self.knot_vector) == len(control_points) + self.degree + 1

    def to_iges(self, *args, **kwargs) -> aerocaps.iges.entity.IGESEntity:
        return aerocaps.iges.curves.RationalBSplineCurveIGES(
            knots=self.knot_vector,
            weights=self.weights,
            control_points_XYZ=self.get_control_point_array(),
            degree=self.degree
        )

    def reverse(self) -> "RationalBezierCurve3D":
        return self.__class__(self.control_points[::-1],
                              self.weights[::-1])

    def evaluate_ndarray(self, t: float) -> np.ndarray:
        """
        Evaluate the NURBS curve at parameter t
        """
        return self.evaluate_simple(t).as_array()

    def get_control_point_array(self) -> np.ndarray:
        return np.array([p.as_array() for p in self.control_points])

    @classmethod
    def generate_from_array(cls, P: np.ndarray, weights: np.ndarray):
        return cls([Point3D(x=Length(m=xyz[0]), y=Length(m=xyz[1]), z=Length(m=xyz[2])) for xyz in P], weights)

    @staticmethod
    def bernstein_poly(n: int, i: int, t: int or float or np.ndarray):
        r"""
        Calculates the Bernstein polynomial for a given Bézier curve order, index, and parameter vector. The
        Bernstein polynomial is described by

        .. math::

            B_{i,n}(t)={n \choose i} t^i (1-t)^{n-i}

        Arguments
        =========
        n: int
            Bézier curve degree (one less than the number of control points in the Bézier curve)
        i: int
            Bézier curve index
        t: int, float, or np.ndarray
            Parameter vector for the Bézier curve

        Returns
        =======
        np.ndarray
            Array of values of the Bernstein polynomial evaluated for each point in the parameter vector
        """
        return nchoosek(n, i) * t ** i * (1.0 - t) ** (n - i)

    def evaluate_simple(self, t: float) -> Point3D:
        n_ctrl_points = len(self.control_points)
        degree = n_ctrl_points - 1
        P = self.get_control_point_array()

        # Evaluate the curve
        w_sum = sum([self.bernstein_poly(degree, i, t) * self.weights[i] for i in range(n_ctrl_points)])
        x = sum([P[i, 0] * self.bernstein_poly(degree, i, t) * self.weights[i] for i in range(n_ctrl_points)])
        y = sum([P[i, 1] * self.bernstein_poly(degree, i, t) * self.weights[i] for i in range(n_ctrl_points)])
        z = sum([P[i, 2] * self.bernstein_poly(degree, i, t) * self.weights[i] for i in range(n_ctrl_points)])

        return Point3D(x=Length(m=x / w_sum), y=Length(m=y / w_sum), z=Length(m=z / w_sum))
        # return Point3D(x=Length(m=x), y=Length(m=y), z=Length(m=z))

    def evaluate(self, t_vec: np.ndarray) -> np.ndarray:
        """
        Evaluates the NURBS curve at a vector of parameter values
        """
        points = np.array([self.evaluate_ndarray(t) for t in t_vec])
        return points

    def plot(self, ax: plt.Axes, projection: str = None, t_vec: np.ndarray = None, **plt_kwargs):
        projection = "XYZ" if projection is None else projection
        t_vec = np.linspace(0.0, 1.0, 201) if t_vec is None else None
        data = self.evaluate(t_vec)
        args = tuple([data[:, self.projection_dict[axis]] for axis in projection])
        ax.plot(*args, **plt_kwargs)

    def plot_control_points(self, ax: plt.Axes, projection: str = None, **plt_kwargs):
        projection = "XYZ" if projection is None else projection
        cps = self.get_control_point_array()
        args = tuple([cps[:, self.projection_dict[axis]] for axis in projection])
        ax.plot(*args, **plt_kwargs)

    def compute_curvature_at_t0(self) -> float:
        """
        Computes the curvature at :math:`t=0` according to Eq. (20) in
        M. S. Floater, “Derivatives of Rational Bézier Curves,” Computer Aided Geometric Design, vol. 9, no. 3,
        pp. 161–174, Aug. 1992, issn: 0167-8396. doi: 10.1016/0167-8396(92)90014-G.

        Returns
        -------
        float
            The curvature at :math:`t=0`
        """
        R1 = (self.weights[0] * self.weights[2]) / self.weights[1]**2
        P01 = Vector3D(p0=self.control_points[0], p1=self.control_points[1])
        P12 = Vector3D(p0=self.control_points[1], p1=self.control_points[2])
        return (self.degree - 1) / self.degree * R1 * P01.cross(P12).mag().m / P01.mag().m**3

    def compute_curvature_at_t1(self) -> float:
        """
        Computes the curvature at :math:`t=1` according to Eq. (20) in
        M. S. Floater, “Derivatives of Rational Bézier Curves,” Computer Aided Geometric Design, vol. 9, no. 3,
        pp. 161–174, Aug. 1992, issn: 0167-8396. doi: 10.1016/0167-8396(92)90014-G.

        Returns
        -------
        float
            The curvature at :math:`t=1`
        """
        R1 = (self.weights[-1] * self.weights[-3]) / self.weights[-2] ** 2
        P_nm2_nm1 = Vector3D(p0=self.control_points[-3], p1=self.control_points[-2])
        P_nm1_n = Vector3D(p0=self.control_points[-2], p1=self.control_points[-1])
        return (self.degree - 1) / self.degree * R1 * P_nm2_nm1.cross(P_nm1_n).mag().m / P_nm1_n.mag().m ** 3

    def enforce_g0(self, other: "RationalBezierCurve3D"):
        other.control_points[0] = self.control_points[-1]

    def enforce_c0(self, other: "RationalBezierCurve3D"):
        self.enforce_g0(other)

    def enforce_g0g1(self, other: "RationalBezierCurve3D", f: float):
        self.enforce_g0(other)
        n_ratio = self.degree / other.degree
        w_ratio_a = self.weights[-2] / self.weights[-1]
        w_ratio_b = other.weights[0] / other.weights[1]
        other.control_points[1] = other.control_points[0] + f * n_ratio * w_ratio_a * w_ratio_b * (self.control_points[-1] - self.control_points[-2])

    def enforce_c0c1(self, other: "RationalBezierCurve3D"):
        self.enforce_g0g1(other, f=1.0)

    def enforce_g0g1g2(self, other: "RationalBezierCurve3D", f: float):
        self.enforce_g0g1(other, f)
        n_ratio_1 = self.degree / other.degree
        n_ratio_2 = (self.degree - 1) / (other.degree - 1)
        n_ratio_3 = 1 / (other.degree - 1)
        w_ratio_1 = self.weights[-3] / self.weights[-1]
        w_ratio_2 = other.weights[0] / other.weights[2]
        w_ratio_3 = self.weights[-2] / self.weights[-1]
        w_ratio_4 = other.weights[1] / other.weights[0]
        other.control_points[2] = other.control_points[1] + f ** 2 * n_ratio_1 * n_ratio_2 * w_ratio_1 * w_ratio_2 * (
                self.control_points[-3] - self.control_points[-2]) - f ** 2 * n_ratio_1 * n_ratio_3 * w_ratio_2 * (
                2 * self.degree * w_ratio_3**2 - (self.degree - 1) * w_ratio_1 - 2 * w_ratio_3) * (
                                          self.control_points[-2] - self.control_points[-1]) + n_ratio_3 * w_ratio_2 * (
                2 * other.degree * w_ratio_4**2 - (other.degree - 1) * w_ratio_2**-1 - 2 * w_ratio_4) * (
                                          other.control_points[1] - other.control_points[0])

    def enforce_c0c1c2(self, other: "RationalBezierCurve3D"):
        self.enforce_g0g1g2(other, f=1.0)


class BSpline3D(Geometry3D):
    def __init__(self,
                 control_points: np.ndarray,
                 knot_vector: np.ndarray,
                 degree: int):
        """
        Non-uniform rational B-spline (NURBS) curve evaluation class
        """
        assert control_points.ndim == 2
        assert knot_vector.ndim == 1
        assert len(knot_vector) == len(control_points) + degree + 1

        self.control_points = control_points
        self.dim = self.control_points.shape[1]
        self.knot_vector = np.array(knot_vector)
        self.weights = np.ones(self.control_points[:, 0].shape)
        self.degree = degree
        self.possible_spans, self.possible_span_indices = self._get_possible_spans()

    def to_iges(self, *args, **kwargs) -> aerocaps.iges.entity.IGESEntity:
        return aerocaps.iges.curves.RationalBSplineCurveIGES(
            knots=self.knot_vector,
            weights=self.weights,
            control_points_XYZ=self.control_points,
            degree=self.degree
        )

    def reverse(self) -> "BSpline3D":
        return self.__class__(np.flipud(self.control_points),
                              (1.0 - self.knot_vector)[::-1],
                              self.degree)

    def evaluate_ndarray(self, t: float) -> np.ndarray:
        """
        Evaluate the NURBS curve at parameter t
        """
        B = self._basis_functions(t, self.degree)
        point = np.dot(B, self.control_points)
        return point

    def evaluate_simple(self, t: float) -> Point3D:
        """
        Evaluates the NURBS curve at parameter t as a ``Point3D``

        Parameters
        ----------
        t: float
            Parameter value

        Returns
        -------
        Point3D
            Value of the NURBS curve at the specified parameter value
        """
        return Point3D.from_array(self.evaluate_ndarray(t))

    def evaluate(self, t_vec: np.ndarray) -> np.ndarray:
        """
        Evaluates the NURBS curve at a vector of parameter values
        """
        points = np.array([self.evaluate_ndarray(t) for t in t_vec])
        return points

    def _get_possible_spans(self) -> (np.ndarray, np.ndarray):
        possible_span_indices = np.array([], dtype=int)
        possible_spans = []
        for knot_idx, (knot_1, knot_2) in enumerate(zip(self.knot_vector[:-1], self.knot_vector[1:])):
            if knot_1 == knot_2:
                continue
            possible_span_indices = np.append(possible_span_indices, knot_idx)
            possible_spans.append([knot_1, knot_2])
        return np.array(possible_spans), possible_span_indices

    def _cox_de_boor(self, t: float, i: int, p: int) -> float:
        if p == 0:
            return 1.0 if i in self.possible_span_indices and self._find_span(t) == i else 0.0
        else:
            with np.errstate(divide="ignore", invalid="ignore"):
                f = (t - self.knot_vector[i]) / (self.knot_vector[i + p] - self.knot_vector[i])
                g = (self.knot_vector[i + p + 1] - t) / (self.knot_vector[i + p + 1] - self.knot_vector[i + 1])
                if np.isinf(f) or np.isnan(f):
                    f = 0.0
                if np.isinf(g) or np.isnan(g):
                    g = 0.0
                if f == 0.0 and g == 0.0:
                    return 0.0
                elif f != 0.0 and g == 0.0:
                    return f * self._cox_de_boor(t, i, p - 1)
                elif f == 0.0 and g != 0.0:
                    return g * self._cox_de_boor(t, i + 1, p - 1)
                else:
                    return f * self._cox_de_boor(t, i, p - 1) + g * self._cox_de_boor(t, i + 1, p - 1)

    def _basis_functions(self, t: float, p: int):
        """
        Compute the non-zero basis functions at parameter t
        """
        return np.array([self._cox_de_boor(t, i, p) for i in range(self.control_points.shape[0])])

    def _find_span(self, t: float):
        """
        Find the knot span index
        """
        # insertion_point = bisect.bisect_left(self.non_repeated_knots, t)
        # return self.possible_spans[insertion_point - 1]
        for knot_span, knot_span_idx in zip(self.possible_spans, self.possible_span_indices):
            if knot_span[0] <= t < knot_span[1]:
                return knot_span_idx
        if t == self.possible_spans[-1][1]:
            return self.possible_span_indices[-1]


class CompositeCurve3D(Geometry3D):
    def __init__(self, curves: typing.List[PCurve3D]):
        self.curves = curves

    def to_iges(self, curve_iges_entities: typing.List[aerocaps.iges.entity.IGESEntity],
                *args, **kwargs) -> aerocaps.iges.entity.IGESEntity:
        return aerocaps.iges.curves.CompositeCurveIGES(
            curve_iges_entities
        )


class CurveOnParametricSurface(Geometry3D):
    def __init__(self,
                 surface: aerocaps.geom.Surface,
                 parametric_curve: aerocaps.geom.Geometry3D,
                 model_space_curve: aerocaps.geom.Geometry3D):
        self.surface = surface
        self.parametric_curve = parametric_curve
        self.model_space_curve = model_space_curve

    def to_iges(self,
                surface_iges: aerocaps.iges.entity.IGESEntity,
                parametric_curve: aerocaps.iges.entity.IGESEntity,
                model_space_curve: aerocaps.iges.entity.IGESEntity,
                *args, **kwargs) -> aerocaps.iges.entity.IGESEntity:
        return aerocaps.iges.curves.CurveOnParametricSurfaceIGES(
            surface_iges,
            parametric_curve,
            model_space_curve
        )


def main():
    bspline = BSpline3D(np.array([
        [1.0, 0.05, 0.0],
        [0.8, 0.12, 0.0],
        [0.6, 0.2, 0.0],
        [0.2, 0.3, 0.0],
        [0.0, 0.05, 0.0],
        [0.0, -0.1, 0.0],
        [0.4, -0.4, 0.0],
        [0.6, -0.05, 0.0],
        [1.0, -0.05, 0.0]
    ]), knot_vector=np.array([0.0, 0.0, 0.0, 0.0, 0.2, 0.375, 0.5, 0.5, 0.75, 1.0, 1.0, 1.0, 1.0]),
        degree=3
    )
    data = bspline.evaluate(np.linspace(0.0, 1.0, 301))
    plt.plot(data[:, 0], data[:, 1], color="steelblue")
    plt.plot(bspline.control_points[:, 0], bspline.control_points[:, 1], ls=":", color="grey", marker="o", mec="steelblue", mfc="none")
    plt.plot([data[75, 0], data[150, 0], data[225, 0]], [data[75, 1], data[150, 1], data[225, 1]], ls="none", marker="o", mfc="indianred", mec="indianred")
    plt.show()


if __name__ == "__main__":
    main()



import os

from .geom.curves import *
from .geom.intersection import *
from .geom.plane import *
from .geom.point import *
from .geom.surfaces import *
from .geom.tools import *
from .geom.transformation import *
from .geom.vector import *
from .units.area import *
from .units.length import *
from .units.angle import *

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_DIR = os.path.join(BASE_DIR, "tests")



from copy import deepcopy

import numpy as np
import pyvista as pv

from aerocaps.units.length import Length
from aerocaps.geom import Geometry2D, Geometry3D
import aerocaps.iges.entity
import aerocaps.iges.point


__all__ = [
    "Point2D",
    "Point3D",
    "Origin2D",
    "Origin3D"
]


class Point2D(Geometry2D):
    def __init__(self, x: Length, y: Length):
        self.x = x
        self.y = y

    def as_array(self, unit: str = "m"):
        return np.array([getattr(self.x, unit), getattr(self.y, unit)])

    @classmethod
    def from_array(cls, arr: np.ndarray, unit: str = "m"):
        return cls(x=Length(**{unit: arr[0]}), y=Length(**{unit: arr[1]}))

    def __add__(self, other):
        return Point2D(x=self.x + other.x, y=self.y + other.y)

    def __sub__(self, other):
        return Point2D(x=self.x - other.x, y=self.y - other.y)

    def __mul__(self, other):
        if isinstance(other, float):
            return Point2D(x=self.x * other, y=self.y * other)
        else:
            raise ValueError("Only multiplication between points and scalars is currently supported")

    def __rmul__(self, other):
        return self.__mul__(other)


class Point3D(Geometry3D):
    def __init__(self, x: Length, y: Length, z: Length):
        self.x = x
        self.y = y
        self.z = z

    def to_iges(self, *args, **kwargs) -> aerocaps.iges.entity.IGESEntity:
        return aerocaps.iges.point.PointIGES(self.as_array())

    def projection_on_principal_plane(self, plane: str = "XY") -> Point2D:
        if plane == "XY":
            return Point2D(x=deepcopy(self.x), y=deepcopy(self.y))
        elif plane == "XZ":
            return Point2D(x=deepcopy(self.x), y=deepcopy(self.z))
        elif plane == "YZ":
            return Point2D(x=deepcopy(self.y), y=deepcopy(self.z))
        else:
            raise ValueError("Invalid plane")

    def as_array(self, unit: str = "m"):
        return np.array([getattr(self.x, unit), getattr(self.y, unit), getattr(self.z, unit)])

    @classmethod
    def from_array(cls, arr: np.ndarray, unit: str = "m"):
        return cls(x=Length(**{unit: arr[0]}), y=Length(**{unit: arr[1]}), z=Length(**{unit: arr[2]}))

    def almost_equals(self, other: "Point3D"):
        return all([np.isclose(self_xyz, other_xyz) for self_xyz, other_xyz in zip(self.as_array(), other.as_array())])

    def plot(self, plot: pv.Plotter, **point_kwargs):
        plot.add_points(np.array([self.as_array()]), **point_kwargs)

    def __add__(self, other):
        return Point3D(x=self.x + other.x, y=self.y + other.y, z=self.z + other.z)

    def __sub__(self, other):
        return Point3D(x=self.x - other.x, y=self.y - other.y, z=self.z - other.z)

    def __mul__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return Point3D(x=self.x * other, y=self.y * other, z=self.z * other)
        else:
            raise ValueError("Only multiplication between points and scalars is currently supported")

    def __rmul__(self, other):
        return self.__mul__(other)


class Origin2D(Point2D):
    def __init__(self):
        super().__init__(x=Length(m=0), y=Length(m=0))


class Origin3D(Point3D):
    def __init__(self):
        super().__init__(x=Length(m=0), y=Length(m=0), z=Length(m=0))



import numpy as np

from aerocaps.iges.iges_generator import IGESGenerator
from aerocaps.iges.curves import LineIGES, BezierIGES, BoundaryCurveIGES, CircularArcIGES
from aerocaps.iges.surfaces import RuledSurfaceIGES, BoundedSurfaceIGES
from aerocaps.iges.transformation import TransformationMatrixIGES
from aerocaps.units.angle import Angle
from aerocaps.units.length import Length


def generate_planar_surface_iges(file_name: str):
    lower_line = LineIGES(np.array([0.0, 0.0, 0.0]), np.array([1.0, 0.0, 0.0]))
    upper_line = LineIGES(np.array([0.2, 1.0, 0.0]), np.array([0.8, 1.0, 0.0]))
    planar_surface = RuledSurfaceIGES(lower_line, upper_line)
    iges_generator = IGESGenerator(entities=[lower_line, upper_line, planar_surface], units="meters")
    iges_generator.generate(file_name=file_name)


def generate_bounded_surface_iges(file_name: str):
    lower_line = LineIGES(np.array([0.0, 0.0, 0.0]), np.array([1.0, 0.0, 0.0]))
    upper_line = LineIGES(np.array([0.2, 1.0, 0.0]), np.array([0.8, 1.0, 0.0]))
    planar_surface = RuledSurfaceIGES(lower_line, upper_line)
    bez_upper = BezierIGES(np.array([
        [0.3, 0.2, 0.0],
        [0.3, 0.25, 0.0],
        [0.35, 0.25, 0.0],
        [0.65, 0.24, 0.0],
        [0.7, 0.2, 0.0]
    ]))
    bez_lower = BezierIGES(np.array([
        [0.3, 0.2, 0.0],
        [0.3, 0.15, 0.0],
        [0.35, 0.15, 0.0],
        [0.65, 0.22, 0.0],
        [0.7, 0.2, 0.0]
    ]))
    outer_circle = CircularArcIGES(Length(m=0.3), start_angle=Angle(deg=0.0), end_angle=Angle(deg=0.0))
    transformation_matrix = TransformationMatrixIGES(tx=Length(m=0.5), ty=Length(m=0.35))
    outer_circle.transformation_matrix.value = transformation_matrix
    boundary_curve_inner = BoundaryCurveIGES(planar_surface, {bez_upper: [], bez_lower: []})
    boundary_curve_outer = BoundaryCurveIGES(planar_surface, {outer_circle: []})
    boundary_surface = BoundedSurfaceIGES(planar_surface, [boundary_curve_inner, boundary_curve_outer])
    entities = [
        lower_line,
        upper_line,
        planar_surface,
        bez_upper,
        bez_lower,
        outer_circle,
        transformation_matrix,
        boundary_curve_inner,
        boundary_curve_outer,
        boundary_surface
    ]
    iges_generator = IGESGenerator(entities=entities, units="meters")
    iges_generator.generate(file_name=file_name)



from abc import abstractmethod

import numpy as np

import aerocaps.iges.entity


class Geometry2D:
    pass


class Geometry3D:
    @abstractmethod
    def to_iges(self, *args, **kwargs) -> aerocaps.iges.entity.IGESEntity:
        """
        Converts the geometric object to an IGES entity. To add this IGES entity to an ``.igs`` file,
        use an :obj:`~aerocaps.iges.iges_generator.IGESGenerator`.
        """
        pass


class Surface(Geometry3D):
    @abstractmethod
    def evaluate(self, Nu: int, Nv: int) -> np.ndarray:
        pass


class InvalidGeometryError(Exception):
    pass


class NegativeWeightError(Exception):
    pass



from aerocaps.units.unit import Unit
from aerocaps.units.area import Area


__all__ = [
    "Length"
]


class Length(Unit):
    """Base-level class for a length dimension with various available units."""
    def __init__(self, ft=None, m=None, inch=None, mm=None, mi=None, nmi=None, km=None, cm=None):
        self._ft, self._m, self._inch, self._mm, self._mi, self._nmi, self._km, self._cm = \
            None, None, None, None, None, None, None, None
        self.convert_from_feet_map = {
            'm': 0.3048,
            'inch': 12,
            'mm': 304.8,
            'mi': 1 / 5280,
            'nmi': 1 / 5280 / 1.150779448,
            'km': 0.0003048,
            'cm': 30.48
        }
        if ft is not None:
            self.ft = ft
        elif m is not None:
            self.m = m
        elif inch is not None:
            self.inch = inch
        elif mm is not None:
            self.mm = mm
        elif mi is not None:
            self.mi = mi
        elif nmi is not None:
            self.nmi = nmi
        elif km is not None:
            self.km = km
        elif cm is not None:
            self.cm = cm
        super().__init__(primary_unit="ft")

    def set_all(self):
        """After setting the length in feet, set the length for all other units"""
        for k, v in self.convert_from_feet_map.items():
            setattr(self, f'_{k}', self._ft * v)

    @property
    def ft(self):
        return self._ft

    @ft.setter
    def ft(self, ft):
        self._ft = ft
        self.set_all()

    @property
    def m(self):
        return self._m

    @m.setter
    def m(self, m):
        self._ft = m / self.convert_from_feet_map['m']
        self.set_all()

    @property
    def inch(self):
        return self._inch

    @inch.setter
    def inch(self, inch):
        self._ft = inch / self.convert_from_feet_map['inch']
        self.set_all()

    @property
    def mm(self):
        return self._mm

    @mm.setter
    def mm(self, mm):
        self._ft = mm / self.convert_from_feet_map['mm']
        self.set_all()

    @property
    def mi(self):
        return self._mi

    @mi.setter
    def mi(self, mi):
        self._ft = mi / self.convert_from_feet_map['mi']
        self.set_all()

    @property
    def nmi(self):
        return self._nmi

    @nmi.setter
    def nmi(self, nmi):
        self._ft = nmi / self.convert_from_feet_map['nmi']
        self.set_all()

    @property
    def km(self):
        return self._km

    @km.setter
    def km(self, km):
        self._ft = km / self.convert_from_feet_map['km']
        self.set_all()

    @property 
    def cm(self): 
        return self._cm
    
    @cm.setter
    def cm(self, cm):
        self._ft = cm / self.convert_from_feet_map['cm']
        self.set_all()

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            new_primary_value = getattr(self, self.primary_unit) * other
            return self.__class__(**{self.primary_unit: new_primary_value})
        elif isinstance(other, Length):
            return Area(m2=self.m * other.m)
        else:
            return NotImplemented



import typing

import numpy as np

from aerocaps.iges.entity import IGESEntity
from aerocaps.iges.iges_param import IGESParam
from aerocaps.units.angle import Angle
from aerocaps.units.length import Length


class CircularArcIGES(IGESEntity):
    """
    IGES IGESEntity #100
    """
    def __init__(self, radius: Length, start_angle: Angle, end_angle: Angle):
        x2 = radius.m * np.cos(start_angle.rad)
        y2 = radius.m * np.sin(start_angle.rad)
        x3 = radius.m * np.cos(end_angle.rad)
        y3 = radius.m * np.sin(end_angle.rad)
        parameter_data = [
            IGESParam(0.0, "real"),
            IGESParam(0.0, "real"),
            IGESParam(0.0, "real"),
            IGESParam(x2, "real"),
            IGESParam(y2, "real"),
            IGESParam(x3, "real"),
            IGESParam(y3, "real")
        ]
        super().__init__(100, parameter_data)


class LineIGES(IGESEntity):
    """
    IGES IGESEntity #110
    """
    def __init__(self, start_point: np.ndarray, end_point: np.ndarray):
        parameter_data = [
            IGESParam(start_point[0], "real"),
            IGESParam(start_point[1], "real"),
            IGESParam(start_point[2], "real"),
            IGESParam(end_point[0], "real"),
            IGESParam(end_point[1], "real"),
            IGESParam(end_point[2], "real")
        ]
        super().__init__(110, parameter_data)


class RationalBSplineCurveIGES(IGESEntity):
    """
    IGES IGESEntity #126
    """
    def __init__(self, knots: np.ndarray, weights: np.ndarray,
                 control_points_XYZ: np.ndarray, degree: int, start_parameter_value=0.0, end_parameter_value=1.0,
                 unit_normal_x=0.0, unit_normal_y=0.0, unit_normal_z=0.0, planar_flag: bool = False,
                 closed_flag: bool = False, polynomial_flag: bool = False, periodic_flag: bool = False,
                 **entity_kwargs):
        self.upper_index_of_sum = control_points_XYZ.shape[0] - 1
        self.degree = degree
        self.flag1 = int(planar_flag)
        self.flag2 = int(closed_flag)
        self.flag3 = int(polynomial_flag)
        self.flag4 = int(periodic_flag)
        self.knots = knots
        self.weights = weights
        self.control_points = control_points_XYZ
        self.v0 = start_parameter_value
        self.v1 = end_parameter_value
        self.XN = unit_normal_x
        self.YN = unit_normal_y
        self.ZN = unit_normal_z
        parameter_data = [
            IGESParam(self.upper_index_of_sum, "int"),
            IGESParam(self.degree, "int"),
            IGESParam(self.flag1, "int"),
            IGESParam(self.flag2, "int"),
            IGESParam(self.flag3, "int"),
            IGESParam(self.flag4, "int"),
            *[IGESParam(k, "real") for k in self.knots],
            *[IGESParam(w, "real") for w in self.weights],
            *[IGESParam(xyz, "real") for xyz in self.control_points.flatten()],
            IGESParam(self.v0, "real"),
            IGESParam(self.v1, "real"),
            IGESParam(self.XN, "real"),
            IGESParam(self.YN, "real"),
            IGESParam(self.ZN, "real"),
        ]
        super().__init__(126, parameter_data, **entity_kwargs)


class BezierIGES(RationalBSplineCurveIGES):
    def __init__(self, control_points_XYZ: np.ndarray, start_parameter_value=0.0, end_parameter_value=1.0,
                 unit_normal_x=0.0, unit_normal_y=0.0, unit_normal_z=0.0, planar_flag: bool = False,
                 closed_flag: bool = False, periodic_flag: bool = False, **entity_kwargs):
        order = len(control_points_XYZ)
        degree = order - 1
        knots = np.concatenate((np.zeros(order), np.ones(order)))
        weights = np.ones(order)
        polynomial_flag = True
        super().__init__(knots=knots, weights=weights, control_points_XYZ=control_points_XYZ, degree=degree,
                         start_parameter_value=start_parameter_value, end_parameter_value=end_parameter_value,
                         unit_normal_x=unit_normal_x, unit_normal_y=unit_normal_y, unit_normal_z=unit_normal_z,
                         planar_flag=planar_flag, closed_flag=closed_flag, polynomial_flag=polynomial_flag,
                         periodic_flag=periodic_flag, **entity_kwargs)


class BoundaryCurveIGES(IGESEntity):
    def __init__(self,
                 untrimmed_surface: IGESEntity,
                 curves: typing.Dict[IGESEntity, typing.List[IGESEntity]],
                 preferred_representation: int = 3,
                 curves_needing_reversal: typing.List[int] = None,
                 **entity_kwargs
                 ):
        """
        IGES Type 141
        """
        if curves_needing_reversal is None:
            curves_needing_reversal = []
        number_parameter_space_curves = sum([len(parameter_space_curves) for parameter_space_curves in curves.values()])
        parameter_data = [
            IGESParam(0 if number_parameter_space_curves == 0 else 1, "int"),
            IGESParam(preferred_representation, "int"),  # 0 = unspecified, 1 = model space, 2 = parameter space,
                                                         # 3 = representations are of equal preference
            IGESParam(untrimmed_surface, "pointer"),
            IGESParam(len(curves) + number_parameter_space_curves, "int")
        ]
        for curve_idx, (model_space_curve, parameter_space_curves) in enumerate(curves.items()):
            parameter_data.append(IGESParam(model_space_curve, "pointer"))
            parameter_data.append(IGESParam(2 if curve_idx in curves_needing_reversal else 1, "int"))
            parameter_data.append(IGESParam(len(parameter_space_curves), "int"))
            for parameter_space_curve in parameter_space_curves:
                parameter_data.append(IGESParam(parameter_space_curve, "pointer"))

        super().__init__(141, parameter_data, **entity_kwargs)


class CurveOnParametricSurfaceIGES(IGESEntity):
    def __init__(self, surface: IGESEntity, parametric_curve: IGESEntity, model_space_curve: IGESEntity, **entity_kwargs):
        """
        IGES Type 142
        """
        parametric_curve.status_number.value = 5
        parameter_data = [
            IGESParam(0, "int"),
            IGESParam(surface, "pointer"),
            IGESParam(parametric_curve, "pointer"),
            IGESParam(model_space_curve, "pointer"),
            IGESParam(3, "int")
        ]
        super().__init__(142, parameter_data, **entity_kwargs)


class CompositeCurveIGES(IGESEntity):
    def __init__(self, curves: typing.List[IGESEntity], **entity_kwargs):
        """
        IGES Type 102
        """
        parameter_data = [
            IGESParam(len(curves), "int"),
            *[IGESParam(curve, "pointer") for curve in curves]
        ]
        super().__init__(102, parameter_data, **entity_kwargs)



import os.path
import typing

from aerocaps.geom.intersection import intersection_of_line_and_plane
from aerocaps.geom.curves import Line3D
from aerocaps.geom.plane import Plane
import aerocaps.iges.curves
from aerocaps.iges.start_end_section import StartSection, EndSection
from aerocaps.iges.global_params import GlobalParams
from aerocaps.iges.entity import IGESEntity, MultiEntityContainer


class IGESGenerator:
    def __init__(self, entities: typing.List[IGESEntity], units: str):
        """
        Creation class for an IGES file using a list of IGES entities. To generate the file, use the
        :obj:`~aerocaps.iges.iges_generator.IGESGenerator.generate` method.

        Parameters
        ----------
        entities: typing.List[IGESEntity]
            List of IGES entities from which a file can be created
        units: str
            Output units to be stored in the IGES file
        """
        self.entities = entities
        self.start = StartSection()
        self.globals = GlobalParams(units)
        self.entity_container = MultiEntityContainer(self.entities)
        self.end_section = None

    def _assign_integer_values_to_pointers(self):
        for entity in self.entities:
            for iges_param in entity.parameter_data:
                if not iges_param.dtype == "pointer":
                    continue
                if not isinstance(iges_param.value, IGESEntity):
                    continue
                iges_param.value = 1 + 2 * self.entities.index(iges_param.value)
            if entity.transformation_matrix.value != 0:
                if not isinstance(entity, IGESEntity):
                    continue
                entity.transformation_matrix.value = 1 + 2 * self.entities.index(entity.transformation_matrix.value)

    def generate(self, file_name: str):
        """
        Generates an IGES file containing all the information for the entities.

        Parameters
        ----------
        file_name: str
            File where the IGES data will be saved. If the file name does not end with the ".igs" or ".iges" extension,
            it will be added automatically.

        Returns
        -------
        str
            The IGES data in Python string format
        """
        # Assign integer values to the pointers
        self._assign_integer_values_to_pointers()

        # First, make sure that the entities know which delimiters to use:
        for entity in self.entity_container.entities:
            entity.param_delimiter = self.globals.parameter_delimiter_char.value
            entity.record_delimiter = self.globals.record_delimiter_char.value

        # Write all the section strings:
        start_section_string = self.start.write_start_section_string()
        global_section_string = self.globals.write_globals_string()
        entity_section_string, data_section_string = self.entity_container.write_all_entity_and_data_strings()
        self.end_section = EndSection(n_start_lines=start_section_string.count("\n"),
                                      n_global_lines=global_section_string.count("\n"),
                                      n_entity_lines=entity_section_string.count("\n"),
                                      n_data_lines=data_section_string.count("\n"))
        end_section_string = self.end_section.write_end_section_string()

        # Add all the section strings together:
        iges_string = start_section_string + global_section_string + entity_section_string + \
                      data_section_string + end_section_string

        # If the file name does not end in the .igs or .iges extension, add the extension:
        if os.path.splitext(file_name)[-1] not in [".igs", ".iges"]:
            file_name += ".igs"

        # Write the total string to the IGES file:
        with open(file_name, "w") as f:
            f.write(iges_string)

        return iges_string
